/*
 * cal.h
 *
 * CPU Abstraction Layer interface.
 */

/* $Id: cal.h,v 1.6 1999/01/09 19:47:47 nyef Exp $ */

#ifndef __CAL_H__
#define __CAL_H__

#include "cpudefs.h"

#define CPUT_NONE      0
#define CPUT_MARAT6502 1
#define CPUT_EMU6502   2
#define CPUT_EMU6502A  3
#define CPUT_EMU65C02  4
#define CPUT_EMU6280   5
#define CPUT_MARATZ80  6
#define CPUT_EMUZ80    7

typedef struct cal_cpu *cal_cpu;

typedef unsigned char (*memread_t)(cal_cpu, unsigned short);
typedef void (*memwrite_t)(cal_cpu, unsigned short, unsigned char);

typedef void (*cpureset_t)(cal_cpu);
typedef void (*cpurun_t)(cal_cpu);
typedef void (*cpurunfor_t)(cal_cpu, int);
typedef void (*cpuirq_t)(cal_cpu);
typedef void (*cpunmi_t)(cal_cpu);
typedef void (*cpusetzpage_t)(cal_cpu, void *);
typedef int (*cputimeleft_t)(cal_cpu);
typedef void (*cpusetmmu_t)(cal_cpu, int, int, memread_t *, memwrite_t *);
typedef void (*cpusetiou_t)(cal_cpu, memread_t, memwrite_t);

struct cal_cpu {
    int type; /* CPUT_xxx */
    void *userdata;
    cpureset_t reset;
    cpurun_t run;
    cpurunfor_t runfor;
    cpuirq_t irq;
    cpunmi_t nmi;
    cpusetzpage_t setzpage;
    cputimeleft_t timeleft;
    cpusetmmu_t setmmu;
    cpusetiou_t setiou;
    union {
#ifdef CPU_MARAT6502
	struct M6502 *d_marat6502;
#endif
#ifdef CPU_EMU6502
	struct emu6502_context *d_emu6502;
#endif
#ifdef CPU_EMU6502A
	struct emu6502_context *d_emu6502a;
#endif
#ifdef CPU_EMU65C02
	struct emu6502_context *d_emu65c02;
#endif
#ifdef CPU_EMU6280
	struct emu6280_context *d_emu6280;
#endif
#ifdef CPU_MARATZ80
	struct Z80 *d_maratz80;
#endif
#ifdef CPU_EMUZ80
	struct emuz80_context *d_emuz80;
#endif
    } data;
};

cal_cpu cal_create(int type);

#endif /* __CAL_H__ */

/*
 * $Log: cal.h,v $
 * Revision 1.6  1999/01/09 19:47:47  nyef
 * added 'void *userdata' to struct cal_cpu.
 *
 * Revision 1.5  1999/01/03 02:25:40  nyef
 * added maratz80 support.
 *
 * Revision 1.4  1999/01/02 06:35:46  nyef
 * added io interface for z80 support.
 *
 * Revision 1.3  1999/01/01 07:00:46  nyef
 * added emuz80 support
 *
 * Revision 1.2  1998/12/12 20:35:13  nyef
 * fixed marat6502 support.
 *
 * Revision 1.1  1998/11/29 18:28:34  nyef
 * Initial revision
 *
 */
