/*
 * cd_freebsd.c
 *
 * FreeBSD CDROM interface
 */

/* $Id: cd_freebsd.c,v 1.1 2000/02/12 19:03:28 nyef Exp $ */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdio.h>

#include <sys/cdio.h>

#include "types.h"
#include "cd.h"

#define CDROM_DEVICE "/dev/acd1c"

int pce_cd_drive;

struct ioc_read_toc_single_entry pce_cd_tocentry;

void cd_set_sector_address(u32 sector)
{
    int result;

    result = lseek(pce_cd_drive, (sector) * 2048, SEEK_SET);
    if (result < 0) {
	perror("pce_cd: seeking to CD sector");
    }
}

void cd_read_next_sector(u8 *buf)
{
    int result;

    result = read(pce_cd_drive, buf, 2048);
    if (result < 0) {
	perror("pce_cd: reading CD data");
    }
}

void cd_get_tocheader(u8 *start_track, u8 *stop_track)
{
    /* pretend that there are 4 tracks on this CD */
    /* FIXME: real implementation please? */
    *start_track = 1;
    *stop_track = 4;
}

void cd_get_tocentry(u8 track, u32 *lba, u8 *flags)
{
    int result;
    
    pce_cd_tocentry.track = track;
    pce_cd_tocentry.address_format = CD_LBA_FORMAT;
    
    result = ioctl(pce_cd_drive, CDIOREADTOCENTRY, &pce_cd_tocentry);
    if (result < 0) {
	perror("pce_cd: reading tocentry");
/*  	deb_printf("pce_cd: tocread failure.\n"); */
    } else {
	*lba = ntohl(pce_cd_tocentry.entry.addr.lba);
	*flags = pce_cd_tocentry.entry.control; /* FIXME: may want cdte.adr << 4 in here as well */
    }
}

void cd_init(void)
{
    pce_cd_drive = open(CDROM_DEVICE, O_RDONLY);
    if (pce_cd_drive < 0) {
	perror("pce_cd: opening CD drive");
	pce_cd_drive = 0;
    } else {
	/* Do something? */
    }
}

/*
 * $Log: cd_freebsd.c,v $
 * Revision 1.1  2000/02/12 19:03:28  nyef
 * Initial revision
 *
 */
