/*
 * genesis.c
 *
 * Sega Genesis emulation.
 */

/* $Id: genesis.c,v 1.4 1999/11/20 05:27:37 nyef Exp $ */

#include <stdlib.h>
#include <netinet/in.h>

#include "ui.h"
#include "system.h"
#include "cal.h"
#include "tool.h"
#include "gen_vdp.h"
#include "emu68k.h"

#define GENESIS_RAMSIZE 0x10000 /* FIXME: Get real ram size data */

/* #define DIVERGANCE_CHECK */

unsigned char *genesis_ram;
cal_cpu genesis_68k;
cal_cpu genesis_z80;
gen_vdp genesis_vdp;
unsigned char *genesis_rom;

/* CPU Cross-verification data */

#ifdef DIVERGANCE_CHECK
cal_cpu genesis_68k_2;

struct memory_event {
    unsigned long addr;
    union {
	unsigned char byte;
	unsigned short word;
    } data;
    int size;
};

struct memory_event divergance[80];
int divergance_count1;
int divergance_count2;
int divergance_phase;
unsigned short last_opword;
int instr_count;
#endif

/*
 * memory map:
 *
 * 000000-3fffff ROM
 * 400000-9fffff Reserved
 * a00000-afffff Z80/Sound
 * b00000-bfffff Reserved
 * c00000-dfffff VDP
 * e00000-feffff RAM Mirrors
 * ff0000-ffffff RAM
 */

unsigned char gen_read8_rom(cal_cpu cpu, unsigned long address)
{
    /* FIXME: range check? */
    return genesis_rom[address & 0x3fffff];
}

unsigned short gen_read16_rom(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_rom: CPU BUG: non-aligned 16-bit access.\n");
    }
    /* FIXME: range check? */
    return htons(*((unsigned short *)(genesis_rom + address)));
}

unsigned char gen_read8_reserved(cal_cpu cpu, unsigned long address)
{
    deb_printf("gen_read8_reserved: address 0x%06x.\n", address);
    return 0;
}

unsigned short gen_read16_reserved(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_reserved: CPU BUG: non-aligned 16-bit access.\n");
    }
    deb_printf("gen_read16_reserved: address 0x%06x.\n", address);
    return 0;
}

unsigned char gen_read8_misc(cal_cpu cpu, unsigned long address)
{
    deb_printf("gen_read8_misc: 0x%06x.\n", address);
    return 0;
}

unsigned short gen_read16_misc(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_misc: CPU BUG: non-aligned 16-bit access.\n");
    }
    deb_printf("gen_read16_misc: 0x%06x.\n", address);
    return 0;
}

unsigned char gen_read8_vdp(cal_cpu cpu, unsigned long address)
{
    deb_printf("gen_read8_vdp: 0x%06x.\n", address);
    return 0;
}

unsigned short gen_read16_vdp(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_vdp: CPU BUG: non-aligned 16-bit access.\n");
    }
    deb_printf("gen_read16_vdp: 0x%06x.\n", address);
    return 0;
}

unsigned char gen_read8_ram(cal_cpu cpu, unsigned long address)
{
    return genesis_ram[address & 0xffff];
}

unsigned short gen_read16_ram(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_ram: CPU BUG: non-aligned 16-bit access.\n");
    }
    return htons(*((unsigned short *)(genesis_ram + (address & 0xffff))));
}

void gen_write8_reserved(cal_cpu cpu, unsigned long address, unsigned char data)
{
    deb_printf("gen_write8_reserved: address 0x%06x, data 0x%02x.\n", address, data);
}

void gen_write16_reserved(cal_cpu cpu, unsigned long address, unsigned short data)
{
    if (address & 1) {
	deb_printf("gen_write16_reserved: CPU BUG: non-aligned 16-bit access.\n");
    }
    deb_printf("gen_write16_reserved: address 0x%06x, data 0x%04hx.\n", address, data);
}

void gen_write8_misc(cal_cpu cpu, unsigned long address, unsigned char data)
{
/*     deb_printf("gen_write8_misc: address 0x%06x, data 0x%02x.\n", address, data); */
}

void gen_write16_misc(cal_cpu cpu, unsigned long address, unsigned short data)
{
    if (address & 1) {
	deb_printf("gen_write16_misc: CPU BUG: non-aligned 16-bit access.\n");
    }
/*     deb_printf("gen_write16_misc: address 0x%06x, data 0x%04hx.\n", address, data); */
}

void gen_write8_vdp(cal_cpu cpu, unsigned long address, unsigned char data)
{
    deb_printf("gen_write8_vdp: address 0x%06x, data 0x%02x.\n", address, data);
}

void gen_write16_vdp(cal_cpu cpu, unsigned long address, unsigned short data)
{
    if (address & 1) {
	deb_printf("gen_write16_vdp: CPU BUG: non-aligned 16-bit access.\n");
    }
/*     deb_printf("gen_write16_vdp: address 0x%06x, data 0x%04hx.\n", address, data); */
    if ((address & 0xfffffc) == 0xc00000) {
	genesis_vdp->writeport0(genesis_vdp, data);
    } else if ((address & 0xfffffc) == 0xc00004) {
	genesis_vdp->writeport1(genesis_vdp, data);
/* 	deb_printf("gen_write16_vdp: CNTRL_LO, data 0x%04hx (%c%c).\n", data, (data & 0x8000)? '1': '0', (data & 0x4000)? '1': '0'); */
/*     } else if (address == 0xc00006) { */
/* 	deb_printf("gen_write16_vdp: CNTRL_HI, data 0x%04hx.\n", data); */
    } else {
	deb_printf("gen_write16_vdp: address 0x%06x, data 0x%04hx.\n", address, data);
    }
}

void gen_write8_ram(cal_cpu cpu, unsigned long address, unsigned char data)
{
    genesis_ram[address & 0xffff] = data;
}

void gen_write16_ram(cal_cpu cpu, unsigned long address, unsigned short data)
{
    if (address & 1) {
	deb_printf("gen_write16_ram: CPU BUG: non-aligned 16-bit access.\n");
    }
    (*((unsigned short *)(genesis_ram + (address & 0xffff)))) = ntohs(data);
}

memread8_t gen_read8[16] = {
    gen_read8_rom,
    gen_read8_rom,
    gen_read8_rom,
    gen_read8_rom,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_misc,
    gen_read8_reserved,
    gen_read8_vdp,
    gen_read8_vdp,
    gen_read8_ram,
    gen_read8_ram,
};
memread16_t gen_read16[16] = {
    gen_read16_rom,
    gen_read16_rom,
    gen_read16_rom,
    gen_read16_rom,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_misc,
    gen_read16_reserved,
    gen_read16_vdp,
    gen_read16_vdp,
    gen_read16_ram,
    gen_read16_ram,
};

memwrite8_t gen_write8[16] = {
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_misc,
    gen_write8_reserved,
    gen_write8_vdp,
    gen_write8_vdp,
    gen_write8_ram,
    gen_write8_ram,
};
memwrite16_t gen_write16[16] = {
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_misc,
    gen_write16_reserved,
    gen_write16_vdp,
    gen_write16_vdp,
    gen_write16_ram,
    gen_write16_ram,
};

#ifdef DIVERGANCE_CHECK
unsigned char gen_div_r8_proc(cal_cpu cpu, unsigned long addr)
{
    unsigned char retval;
    
    if (divergance_phase) {
	if (divergance[divergance_count2].addr != addr) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: invalid address 0x%06lx.\n", divergance_count2, addr);
	}
	if (divergance[divergance_count2].size != 8) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: data size mismatch.\n", divergance_count2);
	}
	retval = divergance[divergance_count2].data.byte;
	divergance_count2++;
    } else {
	divergance[divergance_count1].addr = addr;
	divergance[divergance_count1].size = 8;
	retval = gen_read8[(addr >> 20) & 15](cpu, addr);
	divergance[divergance_count1].data.byte = retval;
	divergance_count1++;
    }
    return retval;
}

void gen_div_w8_proc(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if (divergance_phase) {
	if (divergance[divergance_count2].addr != addr) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: invalid address 0x%06lx.\n", divergance_count2, addr);
	}
	if (divergance[divergance_count2].size != 8) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: data size mismatch.\n", divergance_count2);
	}
	if (divergance[divergance_count2].data.byte != data) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: data value mismatch (0x%02x/0x%02x).\n", divergance_count2, divergance[divergance_count2].data.byte, data);
	}
	divergance_count2++;
    } else {
	divergance[divergance_count1].addr = addr;
	divergance[divergance_count1].size = 8;
	divergance[divergance_count1].data.byte = data;
	gen_write8[(addr >> 20) & 15](cpu, addr, data);
	divergance_count1++;
    }
}

unsigned short gen_div_r16_proc(cal_cpu cpu, unsigned long addr)
{
    unsigned short retval;
    
    if (divergance_phase) {
	if (divergance[divergance_count2].addr != addr) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: invalid address 0x%06lx.\n", divergance_count2, addr);
	}
	if (divergance[divergance_count2].size != 16) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: data size mismatch.\n", divergance_count2);
	}
	retval = divergance[divergance_count2].data.word;
	divergance_count2++;
    } else {
	divergance[divergance_count1].addr = addr;
	divergance[divergance_count1].size = 16;
	retval = gen_read16[(addr >> 20) & 15](cpu, addr);
	divergance[divergance_count1].data.word = retval;
	divergance_count1++;
    }
    return retval;
}

void gen_div_w16_proc(cal_cpu cpu, unsigned long addr, unsigned short data)
{
    if (divergance_phase) {
	if (divergance[divergance_count2].addr != addr) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: invalid address 0x%06lx.\n", divergance_count2, addr);
	}
	if (divergance[divergance_count2].size != 16) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: data size mismatch.\n", divergance_count2);
	}
	if (divergance[divergance_count2].data.word != data) {
	    divergance_phase = 2;
	    system_flags |= F_UNIMPL;
	    deb_printf("divergance at count %d: data value mismatch (0x%04hx/0x%04hx).\n", divergance_count2, divergance[divergance_count2].data.word, data);
	}
	divergance_count2++;
    } else {
	divergance[divergance_count1].addr = addr;
	divergance[divergance_count1].size = 16;
	divergance[divergance_count1].data.word = data;
	gen_write16[(addr >> 20) & 15](cpu, addr, data);
	divergance_count1++;
    }
}

memread8_t gen_div_r8[1] = {
    gen_div_r8_proc,
};

memwrite8_t gen_div_w8[1] = {
    gen_div_w8_proc,
};

memread16_t gen_div_r16[1] = {
    gen_div_r16_proc,
};

memwrite16_t gen_div_w16[1] = {
    gen_div_w16_proc,
};

#endif

void genesis_convert_rom(rom_file romfile)
{
    /* FIXME: kill the person who devised the SMD format */
    unsigned char *newrom;
    unsigned char *cur_ptr;
    int i;
    int j;

    newrom = malloc(romfile->size - 0x200);
    cur_ptr = newrom;
    
    j = 0x200;
    while (j < romfile->size) {
	for (i = 0; i < 0x2000; i++) {
	    *cur_ptr++ = romfile->data[j + 0x2000];
	    *cur_ptr++ = romfile->data[j];
	    j++;
	}
	j += 0x2000;
    }
    free(romfile->data);
    romfile->data = newrom;
}

void genesis_timeslice(cal_cpu cpu)
{
    int i;
    int j;
#ifdef DIVERGANCE_CHECK
    int k;
#endif
    
    for (i = 262; i && (system_flags == F_NONE); i--) {
	cpu->runfor(cpu, 228); /* FIXME: find out how many cycles/hbl */
	for (j = 1; (j < 20) && (!(system_flags & F_UNIMPL)); j++) {
#ifdef DIVERGANCE_CHECK
	    instr_count++;
	    last_opword = divergance[0].data.word;
	    divergance_count1 = 0;
	    divergance_count2 = 0;
	    divergance_phase = 0;
#endif
	    cpu->run(cpu);
#ifdef DIVERGANCE_CHECK
	    divergance_phase = 1;
	    genesis_68k_2->run(genesis_68k_2);
	    for (k = 0; k < 8; k++) {
		if (cpu->data.d_junk68k->regs_a[k] != genesis_68k_2->data.d_junk68k->regs_a[k]) {
		    deb_printf("divergance during sync: a%d mismatch (0x%08lx/0x%08lx).\n", k, cpu->data.d_junk68k->regs_a[k], genesis_68k_2->data.d_junk68k->regs_a[k]);
		    divergance_phase = 2;
		    system_flags |= F_UNIMPL;
		}
		if (cpu->data.d_junk68k->regs_d[k] != genesis_68k_2->data.d_junk68k->regs_d[k]) {
		    deb_printf("divergance during sync: d%d mismatch (0x%08lx/0x%08lx).\n", k, cpu->data.d_junk68k->regs_d[k], genesis_68k_2->data.d_junk68k->regs_d[k]);
		    divergance_phase = 2;
		    system_flags |= F_UNIMPL;
		}
	    }
	    if (cpu->data.d_junk68k->flags != genesis_68k_2->data.d_junk68k->flags) {
		deb_printf("divergance during sync: flag mismatch (0x%04hx/0x%04hx).\n", cpu->data.d_junk68k->flags, genesis_68k_2->data.d_junk68k->flags);
		divergance_phase = 2;
		system_flags |= F_UNIMPL;
	    }
#endif
	}

	if (genesis_vdp->periodic(genesis_vdp)) {
/* 	    cpu->irq(cpu); */
/* 	    deb_printf("gen: VDP wants an IRQ. too bad.\n"); */
	    /* FIXME: Generate IRQ 6 */
#ifdef DIVERGANCE_CHECK
	    divergance_count1 = 0;
	    divergance_count2 = 0;
	    divergance_phase = 0;
	    genesis_68k_2->irq(cpu, 6);
	    divergance_phase = 1;
	    genesis_68k_2->irq(genesis_68k_2, 6);
#else
	    cpu->irq(cpu, 6);
#endif
	}

	if (system_flags & F_UNIMPL) {
#ifdef DIVERGANCE_CHECK
/* 	    if (divergance_phase == 2) { */
		deb_printf("divergance at pc 0x%06lx, opword = 0x%04hx.\n", divergance[0].addr, divergance[0].data.word);
		deb_printf("last opword 0x%04hx.\n", last_opword);
		deb_printf("instr_count: %d.\n", instr_count);
		deb_printf("divergance counts: %d/%d.\n", divergance_count1, divergance_count2);
		for (i = 0; i < divergance_count1; i++) {
		    deb_printf("%d: size %d, addr 0x%06lx, data 0x%04hx.\n", i, divergance[i].size, divergance[i].addr, divergance[i].data.word);
		}
/* 	    } */
#endif
	    deb_printf("Unimplimented Instruction.\n");
	    system_flags |= F_QUIT;
	}
    }
}

void genesis_run(rom_file romfile)
{
    genesis_ram = calloc(1, GENESIS_RAMSIZE);
    
    if (genesis_ram) {
	if (romfile->size & 0x200) {
	    deb_printf("genesis_run: hi/lo paged rom image (SMD format).\n");
	    deb_printf("genesis_run: converting.\n");
	    genesis_convert_rom(romfile);
	} else {
	    deb_printf("genesis_run: linear rom image (BIN format).\n");
	}

	genesis_rom = romfile->data;
	
	genesis_68k = cal_create(CPUT_JUNK68K);
#ifdef DIVERGANCE_CHECK
	genesis_68k_2 = cal_create(CPUT_TORR68K);

	genesis_68k->setmmu8(genesis_68k, 0, 0, gen_div_r8, gen_div_w8);
	genesis_68k->setmmu16(genesis_68k, 0, 0, gen_div_r16, gen_div_w16);
	genesis_68k_2->setmmu8(genesis_68k_2, 0, 0, gen_div_r8, gen_div_w8);
	genesis_68k_2->setmmu16(genesis_68k_2, 0, 0, gen_div_r16, gen_div_w16);
#else
	genesis_68k->setmmu8(genesis_68k, 20, 15, gen_read8, gen_write8);
	genesis_68k->setmmu16(genesis_68k, 20, 15, gen_read16, gen_write16);
#endif
#ifdef DIVERGANCE_CHECK
	divergance_count1 = 0;
	divergance_count2 = 0;
	divergance_phase = 0;
	genesis_68k->reset(genesis_68k);
	divergance_phase = 1;
	genesis_68k_2->reset(genesis_68k_2);

	if (divergance_phase == 2) {
	    deb_printf("divergance during reset.\n");
	}

	instr_count = 0;
#else
	genesis_68k->reset(genesis_68k);
#endif
	
	genesis_vdp = gen_vdp_create();
	
	set_timeslice((void (*)(void *)) &genesis_timeslice, genesis_68k);
    } else {
	deb_printf("out of memory.\n");
    }
}

/*
 * $Log: genesis.c,v $
 * Revision 1.4  1999/11/20 05:27:37  nyef
 * hacked to work with new rom loading interface
 *
 * Revision 1.3  1999/09/08 01:00:39  nyef
 * added preliminary vblank IRQ support (now if I knew it worked...)
 * added some code to compare the behavior of two 68000 cores (disabled)
 *
 * Revision 1.2  1999/08/07 15:38:07  nyef
 * fixed to quit on unimplemented instruction
 *
 * Revision 1.1  1999/04/17 22:24:47  nyef
 * Initial revision
 *
 */
