/*
 * io_nes.S
 *
 * NES I/O space emulation
 */

/* $Id: io_nes.S,v 1.18 1998/12/17 06:12:15 nyef Exp $ */

#include "opc6502.h"
	
	.globl SYMNAME(io_nes_init)
SYMLABEL(io_nes_init)
	pushl %edi

	movl 0x8(%esp), %edi

	cld
	
	movl $nes_r_bank_0, %eax /* system ram */
	stosl
	movl $nes_io_read1, %eax /* IO area */
	stosl
	movl $nes_io_read2, %eax /* IO area */
	stosl
	movl $nes_r_bank_6, %eax /* cart ram */
	stosl
	movl $nes_r_bank_8, %eax /* rom */
	stosl
	movl $nes_r_bank_a, %eax /* rom */
	stosl
	movl $nes_r_bank_c, %eax /* rom */
	stosl
	movl $nes_r_bank_e, %eax /* rom */
	stosl

	movl $nes_w_bank_0, %eax /* system ram */
	stosl
	movl $nes_io_write1, %eax /* IO area */
	stosl
	movl $nes_io_write2, %eax /* IO area */
	stosl
	movl $nes_w_bank_6, %eax /* cart ram */
	stosl
	movl $nes_w_mapper, %eax /* rom area: write to mapper */
	stosl
	stosl
	stosl
	stosl

	popl %edi
	ret

	
/* memory access routines */

	.globl nes_r_memzero
nes_r_memzero:
	movb $0x80, %al
	ret

	.globl nes_r_bank_0
nes_r_bank_0:
	movl 8(%esp), %eax
	pushl %esi
	movl SYMNAME(bank_0), %esi
	andl $0x7ff, %eax
	movb (%esi, %eax), %al
	popl %esi
	ret

	.globl nes_r_bank_6
nes_r_bank_6:
	movl 8(%esp), %eax
	pushl %esi
	movl SYMNAME(bank_6), %esi
	andl $0x1fff, %eax
	movb (%esi, %eax), %al
	popl %esi
	ret

	.globl nes_r_bank_8
nes_r_bank_8:
	movl 8(%esp), %eax
	pushl %esi
	movl SYMNAME(bank_8), %esi
	andl $0x1fff, %eax
	movb (%esi, %eax), %al
	popl %esi
	ret

	.globl nes_r_bank_a
nes_r_bank_a:
	movl 8(%esp), %eax
	pushl %esi
	movl SYMNAME(bank_A), %esi
	andl $0x1fff, %eax
	movb (%esi, %eax), %al
	popl %esi
	ret

	.globl nes_r_bank_c
nes_r_bank_c:
	movl 8(%esp), %eax
	pushl %esi
	movl SYMNAME(bank_C), %esi
	andl $0x1fff, %eax
	movb (%esi, %eax), %al
	popl %esi
	ret

	.globl nes_r_bank_e
nes_r_bank_e:
	movl 8(%esp), %eax
	pushl %esi
	movl SYMNAME(bank_E), %esi
	andl $0x1fff, %eax
	movb (%esi, %eax), %al
	popl %esi
	ret

	.globl nes_w_ignore
nes_w_ignore:	
	ret

	.globl nes_w_bank_0
nes_w_bank_0:
	pushl %eax
	pushl %edi
	movl 20(%esp), %eax
	movl 16(%esp), %edi
	andl $0x7ff, %edi
	addl SYMNAME(bank_0), %edi
	movb %al, (%edi)
	popl %edi
	popl %eax
	ret

	.globl nes_w_bank_6
nes_w_bank_6:
	pushl %eax
	pushl %edi
	movl 20(%esp), %eax
	movl 16(%esp), %edi
	andl $0x1fff, %edi
	addl SYMNAME(bank_6), %edi
	movb %al, (%edi)
	popl %edi
	popl %eax
	ret

	.globl nes_w_mapper
nes_w_mapper:
	pushl %eax
	pushl %edi
	movl 20(%esp), %eax
	movl 16(%esp), %edi
	andl $0xff, %eax
	pushl %eax
	pushl %edi
	call *SYMNAME(cur_write)
	addl $8, %esp
	popl %edi
	popl %eax
	ret

/* I/O emulation routines */
	
io_nes_rtbl1:
	.long SYMNAME(ppu_r_2000)
	.long SYMNAME(ppu_r_2001)
	.long SYMNAME(ppu_r_2002)
	.long SYMNAME(ppu_r_2003)
	.long SYMNAME(ppu_r_2004)
	.long SYMNAME(ppu_r_2005)
	.long SYMNAME(ppu_r_2006)
	.long SYMNAME(ppu_r_2007)

	.globl nes_io_read1
nes_io_read1:
	movl 8(%esp), %eax
	andl $0x7, %eax
	jmp io_nes_rtbl1(,%eax,4)

io_nes_wtbl1:
	.long SYMNAME(ppu_w_2000)
	.long SYMNAME(ppu_w_2001)
	.long SYMNAME(ppu_w_2002)
	.long SYMNAME(ppu_w_2003)
	.long SYMNAME(ppu_w_2004)
	.long SYMNAME(ppu_w_2005)
	.long SYMNAME(ppu_w_2006)
	.long SYMNAME(ppu_w_2007)

	.globl nes_io_write1
nes_io_write1:
	pushl %eax
	movl 12(%esp), %eax
	andl $0x7, %eax
	pushl 16(%esp)
	call io_nes_wtbl1(,%eax,4)
	add $4, %esp
	popl %eax
	ret
	
io_nes_rtbl2:
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long nes_r_memzero
	.long io_nes_r_4016
	.long io_nes_r_4017

	.globl nes_io_read2
nes_io_read2:
	movl 8(%esp), %eax
	andl $0x1fff, %eax
	cmpl $0x17, %eax
	jg 0f
	call io_nes_rtbl2(,%eax,4)
0:	ret

io_nes_wtbl2:
	.long SYMNAME(nes_psg_write_c1a)
	.long SYMNAME(nes_psg_write_c1b)
	.long SYMNAME(nes_psg_write_c1c)
	.long SYMNAME(nes_psg_write_c1d)
	.long SYMNAME(nes_psg_write_c2a)
	.long SYMNAME(nes_psg_write_c2b)
	.long SYMNAME(nes_psg_write_c2c)
	.long SYMNAME(nes_psg_write_c2d)
	.long SYMNAME(nes_psg_write_c3a)
	.long SYMNAME(nes_psg_write_c3b)
	.long SYMNAME(nes_psg_write_c3c)
	.long SYMNAME(nes_psg_write_c3d)
	.long SYMNAME(nes_psg_write_c4a)
	.long SYMNAME(nes_psg_write_c4b)
	.long SYMNAME(nes_psg_write_c4c)
	.long SYMNAME(nes_psg_write_c4d)
	.long nes_w_ignore
	.long nes_w_ignore
	.long nes_w_ignore
	.long nes_w_ignore
	.long io_nes_w_4014
	.long SYMNAME(nes_psg_write_control)
	.long io_nes_w_4016
	.long nes_w_ignore

	.globl nes_io_write2
nes_io_write2:
	pushl %eax
	movl 12(%esp), %eax
	andl $0x1fff, %eax
	cmpl $0x17, %eax
	jg 0f
	movl io_nes_wtbl2(,%eax,4), %eax
	xchg %eax, (%esp)
	ret
0:	popl %eax
	ret

io_nes_w_4014:
	pushl 12(%esp)
	andl $0xff, (%esp)
	call SYMNAME(PPU_do_dma)
	addl $4, %esp
	ret

	.data
	.globl SYMNAME(joy1_data)
SYMLABEL(joy1_data)
	.long 0
joy1_latch:
	.long 0
	.text
	
io_nes_w_4016:
	pushl %esi
	movl SYMNAME(joy1_data), %esi
	movl %esi, joy1_latch
	popl %esi
	ret

io_nes_r_4016:
	movb $0x20, %al
	clc
	rcrl joy1_latch
	rclb %al
	ret

io_nes_r_4017:
	movb $0x40, %al
	ret

/*
 * $Log: io_nes.S,v $
 * Revision 1.18  1998/12/17 06:12:15  nyef
 * added SYMNAME()s to the sound procedures in io_nes_wtbl2.
 *
 * Revision 1.17  1998/12/12 22:09:57  nyef
 * cleaned up some of the code.
 * fixed a bug reading from bank 6.
 *
 * Revision 1.16  1998/12/12 20:32:14  nyef
 * continued converting to the new memory access spec.
 *
 * Revision 1.15  1998/11/29 18:54:09  nyef
 * removed commented-out code, cleaned up io_nes_init().
 * also changed io_nes_w_4014() to be closer to spec.
 *
 * Revision 1.14  1998/11/29 18:27:38  nyef
 * continued converting to the new memory access spec.
 *
 * Revision 1.13  1998/11/29 17:19:20  nyef
 * started to convert to the new memory access spec.
 *
 * Revision 1.12  1998/10/29 01:27:57  nyef
 * moved all PPU I/O emulation routines to nes_ppu.c.
 * left old code in for reference.
 *
 * Revision 1.11  1998/10/22 02:02:38  nyef
 * moved ppu_cntrl1, ppu_cntrl2, ppu_status, ppu_spraddy, ppu_vscrol,
 * ppu_hscrol, ppu_scrollmode, PPU_addr and PPU_amode out to nes_ppu.c
 *
 * Revision 1.10  1998/09/03 01:24:25  nyef
 * added SYMNAMEs around calls to psg routines. oops.
 *
 * Revision 1.9  1998/09/02 01:32:50  nyef
 * changed joypad 2 read (4017) to return 0x40 instead of 0x42.
 * this fixed zelda 1, smb 1.
 *
 * Revision 1.8  1998/08/31 01:51:07  nyef
 * fixed stupid mistake in setting up the psg interface.
 *
 * Revision 1.7  1998/08/30 20:30:49  nyef
 * added psg interface. it should probably be replaced with something
 * a little neater looking in the future, but this works for now.
 *
 * Revision 1.6  1998/08/24 00:31:22  nyef
 * removed old scroll debug message text.
 *
 * Revision 1.5  1998/08/02 19:35:03  nyef
 * squashed v-scroll bug (change to io_nes_w_2005)
 *
 * Revision 1.4  1998/08/02 15:07:52  nyef
 * added simple handlers for unknown I/O access. (just prints a message)
 *
 * Revision 1.3  1998/07/29 01:02:27  nyef
 * changed to use SYMNAME/SYMLABEL for all labels referenced from C.
 *
 * Revision 1.2  1998/07/26 23:15:21  nyef
 * added ZPage write code.
 *
 * Revision 1.1  1998/07/11 22:17:44  nyef
 * Initial revision
 *
 */
