/*
 * mappers.c
 *
 * mapper emulation routines
 */

/* $Id: mappers.c,v 1.36 1998/12/20 03:43:13 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include "mappers.h"
#include "ui.h"
#include "tool.h"
#include "nes_ppu.h"
#include "nes.h"
#include "cal.h"

/* defines for mapper interfaces */
extern mapper_iface_t map0_iface;
extern mapper_iface_t map1_iface;
extern mapper_iface_t map2_iface;
extern mapper_iface_t map3_iface;
extern mapper_iface_t map4_iface;
extern mapper_iface_t map5_iface;
extern mapper_iface_t map6_iface;
extern mapper_iface_t map7_iface;
extern mapper_iface_t map8_iface;
extern mapper_iface_t map9_iface;
extern mapper_iface_t mapA_iface;
extern mapper_iface_t mapB_iface;
extern mapper_iface_t mapC_iface;
extern mapper_iface_t mapD_iface;
extern mapper_iface_t mapE_iface;
extern mapper_iface_t mapF_iface;

/* mapper support. 0: none. 1: partial. 2: full. */
int mapper_support[16] = {2, 2, 2, 2, 1, 0, 0, 2, 0, 2, 2, 2, 0, 0, 0, 0};

/* mapper iface record table */
mapper_iface_t *mapper_ifaces[16] = {
    &map0_iface,
    &map1_iface,
    &map2_iface,
    &map3_iface,
    &map4_iface,
    NULL,
    NULL,
    &map7_iface,
    NULL,
    &map9_iface,
    &mapA_iface,
    &mapB_iface,
    NULL,
    NULL,
    NULL,
    NULL,
};

int cur_mapper = -1;

mapinit_t cur_init;
mapwrite_t cur_write;
maphsync_t cur_hsync;
mapvsync_t cur_vsync;

int set_mapper(int mapper)
{
    if (mapper_support[mapper]) {
	deb_printf("mappers: initializing %s supported mapper %d.\n",
		   (mapper_support[mapper] == 1)? "partially": "completely",
		   mapper);
        cur_mapper = mapper;
	cur_init = mapper_ifaces[mapper]->initcode;
	cur_write = mapper_ifaces[mapper]->writecode;
	cur_hsync = mapper_ifaces[mapper]->hsynccode;
	cur_vsync = mapper_ifaces[mapper]->vsynccode;
        return 0;
    } else {
        return -1;
    }
}

int get_mapper(void)
{
    return cur_mapper;
}

int check_mapper(int mapper)
{
    return mapper_support[mapper];
}


/* dummy routines for when a mapper doesn't need it. */

void dummy_hsync(cal_cpu cpu)
{
}

void dummy_vsync(cal_cpu cpu)
{
}

/* MAPPER 0 Implementation */

void map0_write(unsigned short address, unsigned char value)
{
}

void map0_init(void)
{
    ppu_latchfunc = NULL;
    if (ppu_flags & PPUF_VROM) {
	PPU_bank_00 = &vromimage[0x0000];
	PPU_bank_04 = &vromimage[0x0400];
	PPU_bank_08 = &vromimage[0x0800];
	PPU_bank_0c = &vromimage[0x0c00];
	PPU_bank_10 = &vromimage[0x1000];
	PPU_bank_14 = &vromimage[0x1400];
	PPU_bank_18 = &vromimage[0x1800];
	PPU_bank_1c = &vromimage[0x1c00];
    } else {
	deb_printf("map0_init(): cart has no vrom. not good.\n");
    }
    if (cart_vrom_size > 1) {
	deb_printf("map0_init(): cart has more than 8k vrom. not good.\n");
    }
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map0_iface = {map0_init, map0_write, dummy_hsync, dummy_vsync};


/* MAPPER 1 Implementation */

int map1_seq;
int map1_acc;
int map1_data[4];

void map1_write(unsigned short address, unsigned char value)
{
    int i;
    int bank_select;
    
    if (value & 0x80) {
	map1_data[0] |= 0x0c;
	map1_acc = map1_data[(address >> 13) & 3];
	map1_seq = 5;
    } else {
	map1_acc |= ((value & 1) << (map1_seq));
	map1_seq++;
    }

    if ((map1_seq) == 5) {
	map1_data[(address >> 13) & 3] = map1_acc;
	map1_seq = 0;
	map1_acc = 0;
	    
	if (!(map1_data[0] & 2)) {
	    PPU_mirror_one_low();
	} else if (map1_data[0] & 1) {
	    PPU_mirror_horizontal();
	} else {
	    PPU_mirror_vertical();
	}
	
	if (cart_rom_size == 0x20) { /* 512k cart */
	    bank_select = (map1_data[1] & 0x10) << 1;
	} else if (cart_rom_size == 0x40) { /* 1024k cart */
	    if (map1_data[0] & 0x10) {
		bank_select = ((map1_data[1] & 0x10) << 1) | ((map1_data[2] & 0x10) << 2);
	    } else {
		bank_select = (map1_data[1] & 0x10)? 0x60: 0x00;
	    }
	} else { /* other carts */
	    bank_select = 0;
	}
	
	if (!(map1_data[0] & 8)) {
	    for (i = 0xf; ((map1_data[3]) & i) >= (cart_rom_size >> 1); i >>= 1) {
		/* mask down to the rom size. */
	    }
	    bank8_reg = ((map1_data[3] & i) << 2) + bank_select;
	    bankA_reg = ((map1_data[3] & i) << 2) + bank_select + 1;
	    bankC_reg = ((map1_data[3] & i) << 2) + bank_select + 2;
	    bankE_reg = ((map1_data[3] & i) << 2) + bank_select + 3;
	} else if (map1_data[0] & 4) {
	    for (i = 0xf; ((map1_data[3]) & i) >= cart_rom_size; i >>= 1) {
		/* mask down to the rom size. */
	    }
	    bank8_reg = ((map1_data[3] & i) << 1) + bank_select;
	    bankA_reg = ((map1_data[3] & i) << 1) + bank_select + 1;
	    bankC_reg = ((cart_rom_size - 1) << 1);
	    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
	} else {
	    for (i = 0xf; ((map1_data[3]) & i) >= cart_rom_size; i >>= 1) {
		/* mask down to the rom size. */
	    }
	    bank8_reg = 0;
	    bankA_reg = 1;
	    bankC_reg = ((map1_data[3] & i) << 1) + bank_select;
	    bankE_reg = ((map1_data[3] & i) << 1) + bank_select + 1;
	}
	sync_banks();
	
	if (ppu_flags & PPUF_VROM) {
	    if (map1_data[0] & 0x10) {
		for (i = 0x1f; ((map1_data[1]) & i) >= (cart_vrom_size << 1); i >>= 1) {
		    /* mask down to the vrom size. */
		}
		PPU_bank_00 = &vromimage[(map1_data[1]&i)*0x1000+0x0000];
		PPU_bank_04 = &vromimage[(map1_data[1]&i)*0x1000+0x0400];
		PPU_bank_08 = &vromimage[(map1_data[1]&i)*0x1000+0x0800];
		PPU_bank_0c = &vromimage[(map1_data[1]&i)*0x1000+0x0c00];
		for (i = 0x1f; ((map1_data[2]) & i) >= (cart_vrom_size << 1); i >>= 1) {
		    /* mask down to the vrom size. */
		}
		PPU_bank_10 = &vromimage[(map1_data[2]&i)*0x1000+0x0000];
		PPU_bank_14 = &vromimage[(map1_data[2]&i)*0x1000+0x0400];
		PPU_bank_18 = &vromimage[(map1_data[2]&i)*0x1000+0x0800];
		PPU_bank_1c = &vromimage[(map1_data[2]&i)*0x1000+0x0c00];
	    } else {
		for (i = 0x1f; ((map1_data[1]) & i) >= (cart_vrom_size << 1); i >>= 1) {
		    /* mask down to the vrom size. */
		}
		i &= 0xfe;
		PPU_bank_00 = &vromimage[(map1_data[1]&i)*0x1000+0x0000];
		PPU_bank_04 = &vromimage[(map1_data[1]&i)*0x1000+0x0400];
		PPU_bank_08 = &vromimage[(map1_data[1]&i)*0x1000+0x0800];
		PPU_bank_0c = &vromimage[(map1_data[1]&i)*0x1000+0x0c00];
		PPU_bank_10 = &vromimage[(map1_data[1]&i)*0x1000+0x1000];
		PPU_bank_14 = &vromimage[(map1_data[1]&i)*0x1000+0x1400];
		PPU_bank_18 = &vromimage[(map1_data[1]&i)*0x1000+0x1800];
		PPU_bank_1c = &vromimage[(map1_data[1]&i)*0x1000+0x1c00];
	    }
	}
    }
}

void map1_init(void)
{
    ppu_latchfunc = NULL;
    map1_data[0] = 0x1f; /* FIXME: may be incorrect */
    map1_data[3] = 0x00;
    if (ppu_flags & PPUF_VROM) {
	/* FIXME: may want to set consistent internal state */
	PPU_bank_00 = &vromimage[0x0000];
	PPU_bank_04 = &vromimage[0x0400];
	PPU_bank_08 = &vromimage[0x0800];
	PPU_bank_0c = &vromimage[0x0c00];
	PPU_bank_10 = &vromimage[0x1000];
	PPU_bank_14 = &vromimage[0x1400];
	PPU_bank_18 = &vromimage[0x1800];
	PPU_bank_1c = &vromimage[0x1c00];
    }
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map1_iface = {map1_init, map1_write, dummy_hsync, dummy_vsync};


/* MAPPER 2 Implementation */

void map2_write(unsigned short address, unsigned char value)
{
/*     deb_printf("map2_write(%04hx, %02x)\n", address, value); */
    if (value >= cart_rom_size) {
	int i;
	
	deb_printf("map2_write: page %02x beyond end of cartridge, fudging values.\n", value);

	i = 0x7f;
	while ((value &= i) >= cart_rom_size) {
	    i >>= 1;
	}
	deb_printf("map2_write: substituting page %02x.\n", value);
    }
    bank8_reg = (value << 1);
    bankA_reg = (value << 1) + 1;
    sync_banks();
}

void map2_init(void)
{
    ppu_latchfunc = NULL;
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map2_iface = {map2_init, map2_write, dummy_hsync, dummy_vsync};


/* MAPPER 3 Implementation */

void map3_write(unsigned short address, unsigned char value)
{
/*     deb_printf("map3_write(0x%04hx, 0x%02x)\n", address, value); */
    if (value >= cart_vrom_size) {
	int i;
	
	deb_printf("map3_write: page beyond end of cartridge, fudging values.\n");
	
	i = 0x7f;
	while ((value &= i) >= cart_vrom_size) {
	    i >>= 1;
	}
    }

    PPU_bank_00 = &vromimage[value*0x2000+0x0000];
    PPU_bank_04 = &vromimage[value*0x2000+0x0400];
    PPU_bank_08 = &vromimage[value*0x2000+0x0800];
    PPU_bank_0c = &vromimage[value*0x2000+0x0c00];
    PPU_bank_10 = &vromimage[value*0x2000+0x1000];
    PPU_bank_14 = &vromimage[value*0x2000+0x1400];
    PPU_bank_18 = &vromimage[value*0x2000+0x1800];
    PPU_bank_1c = &vromimage[value*0x2000+0x1c00];
}

void map3_init(void)
{
    ppu_latchfunc = NULL;
    map3_write(0, 0);
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map3_iface = {map3_init, map3_write, dummy_hsync, dummy_vsync};


/* MAPPER 4 Implementation */

unsigned char map4_cmd; /* current mapper command */
unsigned char map4_irq_count;
unsigned char map4_irq_latch;
unsigned char map4_irq_state;

unsigned char map4_vrom_switch[6];
unsigned char map4_rom_switch[2];

void map4_sync_rom(void)
{
    if (map4_cmd & 0x40) {
	bank8_reg = 0; /* FIXME: may be wrong */
	bankA_reg = 1;
	bankC_reg = map4_rom_switch[0];
	bankE_reg = map4_rom_switch[1];
    } else {
	bank8_reg = map4_rom_switch[0];
	bankA_reg = map4_rom_switch[1];
	bankC_reg = ((cart_rom_size - 1) << 1);
	bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    }
    sync_banks();
}

void map4_sync_vrom(void)
{
    if (ppu_flags & PPUF_VROM) {
	if (map4_cmd & 0x80) {
	    PPU_bank_10 = &vromimage[(map4_vrom_switch[0]*0x0400+0x0000)];
	    PPU_bank_14 = &vromimage[(map4_vrom_switch[0]*0x0400+0x0400)];
	    PPU_bank_18 = &vromimage[(map4_vrom_switch[1]*0x0400+0x0000)];
	    PPU_bank_1c = &vromimage[(map4_vrom_switch[1]*0x0400+0x0400)];
	    PPU_bank_00 = &vromimage[(map4_vrom_switch[2]*0x0400+0x0000)];
	    PPU_bank_04 = &vromimage[(map4_vrom_switch[3]*0x0400+0x0000)];
	    PPU_bank_08 = &vromimage[(map4_vrom_switch[4]*0x0400+0x0000)];
	    PPU_bank_0c = &vromimage[(map4_vrom_switch[5]*0x0400+0x0000)];
	} else {
	    PPU_bank_00 = &vromimage[(map4_vrom_switch[0]*0x0400+0x0000)];
	    PPU_bank_04 = &vromimage[(map4_vrom_switch[0]*0x0400+0x0400)];
	    PPU_bank_08 = &vromimage[(map4_vrom_switch[1]*0x0400+0x0000)];
	    PPU_bank_0c = &vromimage[(map4_vrom_switch[1]*0x0400+0x0400)];
	    PPU_bank_10 = &vromimage[(map4_vrom_switch[2]*0x0400+0x0000)];
	    PPU_bank_14 = &vromimage[(map4_vrom_switch[3]*0x0400+0x0000)];
	    PPU_bank_18 = &vromimage[(map4_vrom_switch[4]*0x0400+0x0000)];
	    PPU_bank_1c = &vromimage[(map4_vrom_switch[5]*0x0400+0x0000)];
	}
    }
}

void map4_write(unsigned short address, unsigned char value)
{
/*    deb_printf("map4_write(0x%04hx, 0x%02x)\n", address, value); */
    if (address == 0x8000) {
        map4_cmd = value;
	map4_sync_vrom();
	map4_sync_rom();
    } else if (address == 0x8001) {
        switch (map4_cmd & 7) {
	case 0:
	    map4_vrom_switch[0] = value;
	    break;
	case 1:
	    map4_vrom_switch[1] = value;
	    break;
	case 2:
	    map4_vrom_switch[2] = value;
	    break;
	case 3:
	    map4_vrom_switch[3] = value;
	    break;
	case 4:
	    map4_vrom_switch[4] = value;
	    break;
	case 5:
	    map4_vrom_switch[5] = value;
	    break;
	case 6:
	    if (value > (cart_rom_size << 1)) {
		int i;
		
		i = 0x7f;
		while ((value &= i) >= (cart_rom_size << 1)) {
		    i >>= 1;
		}
	    }
            map4_rom_switch[0] = value;
	    break;
	case 7:
	    if (value > (cart_rom_size << 1)) {
		int i;
		
		i = 0x7f;
		while ((value &= i) >= (cart_rom_size << 1)) {
		    i >>= 1;
		}
	    }
            map4_rom_switch[1] = value;
	    break;
	default:
	    deb_printf("map4_write: unknown command. (0x%02x)\n", map4_cmd);
        }
	map4_sync_rom();
	map4_sync_vrom();
    } else if (address == 0xa000) {
/*         deb_printf("map4_write: mirror control? (0x%02x)\n", value); */
	if (value & 0x40) {
	    PPU_mirror_one_high();
	} else {
	    if (value & 0x01) {
		PPU_mirror_horizontal();
	    } else {
		PPU_mirror_vertical();
	    }
	}
    } else if (address == 0xc000) {
	map4_irq_count = value;
    } else if (address == 0xc001) {
	map4_irq_latch = value;
    } else if (address == 0xe000) {
	map4_irq_state = 0;
	map4_irq_count = map4_irq_latch;
    } else if (address == 0xe001) {
	map4_irq_state = 1;
    } else {
        deb_printf("map4_write: unknown address. (0x%04hx, 0x%02x)\n", address, value);
    }
}

void map4_hsync(cal_cpu cpu)
{
    if (map4_irq_state && (map4_irq_count-- == 0)) {
	map4_irq_count = map4_irq_latch;
	cpu->irq(cpu);
    }
}

void map4_vsync(cal_cpu cpu)
{
	map4_irq_count = map4_irq_latch;
}

void map4_init(void)
{
    ppu_latchfunc = NULL;
    map4_cmd = 0;
    map4_rom_switch[0] = 0;
    map4_rom_switch[1] = 1;
    map4_sync_rom();
    map4_vrom_switch[0] = 0;
    map4_vrom_switch[1] = 2;
    map4_vrom_switch[2] = 4;
    map4_vrom_switch[3] = 5;
    map4_vrom_switch[4] = 6;
    map4_vrom_switch[5] = 7;
    map4_sync_vrom();
}

mapper_iface_t map4_iface = {map4_init, map4_write, map4_hsync, map4_vsync};


/* MAPPER 7 Implementation */

void map7_write(unsigned short address, unsigned char value)
{
    if ((value & 0xf) >= (cart_rom_size >> 1)) {
	int i;

	deb_printf("map7_write: page beyond end of cartridge, fudging values.\n");

	i = 0x7f;
	while ((value &= i) >= cart_vrom_size) {
	    i >>= 1;
	}
    }
    bank8_reg = (value & 0xf) << 2;
    bankA_reg = bank8_reg + 1;
    bankC_reg = bankA_reg + 1;
    bankE_reg = bankC_reg + 1;
    sync_banks();

    if (value & 0x10) {
	PPU_mirror_one_high();
    } else {
	PPU_mirror_one_low();
    }
}

void map7_init(void)
{
    ppu_latchfunc = NULL;
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = 2;
    bankE_reg = 3;
    sync_banks();
}

mapper_iface_t map7_iface = {map7_init, map7_write, dummy_hsync, dummy_vsync};


/* MAPPER 9 Implementation */

unsigned char map9_latch1_lo;
unsigned char map9_latch1_hi;
unsigned char map9_latch1_state;
unsigned char map9_latch2_lo;
unsigned char map9_latch2_hi;
unsigned char map9_latch2_state;

void map9_latchfunc(unsigned short address)
{
    if ((address & 0x3ff0) == 0x0fd0) {
	map9_latch1_state = 0;
	PPU_bank_00 = &vromimage[(map9_latch1_lo << 12) + 0x0000];
	PPU_bank_04 = &vromimage[(map9_latch1_lo << 12) + 0x0400];
	PPU_bank_08 = &vromimage[(map9_latch1_lo << 12) + 0x0800];
	PPU_bank_0c = &vromimage[(map9_latch1_lo << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x0fe0) {
	map9_latch1_state = 1;
	PPU_bank_00 = &vromimage[(map9_latch1_hi << 12) + 0x0000];
	PPU_bank_04 = &vromimage[(map9_latch1_hi << 12) + 0x0400];
	PPU_bank_08 = &vromimage[(map9_latch1_hi << 12) + 0x0800];
	PPU_bank_0c = &vromimage[(map9_latch1_hi << 12) + 0x0c00];
    }
    if ((address & 0x3ff0) == 0x1fd0) {
	map9_latch2_state = 0;
	PPU_bank_10 = &vromimage[(map9_latch2_lo << 12) + 0x0000];
	PPU_bank_14 = &vromimage[(map9_latch2_lo << 12) + 0x0400];
	PPU_bank_18 = &vromimage[(map9_latch2_lo << 12) + 0x0800];
	PPU_bank_1c = &vromimage[(map9_latch2_lo << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x1fe0) {
	map9_latch2_state = 1;
	PPU_bank_10 = &vromimage[(map9_latch2_hi << 12) + 0x0000];
	PPU_bank_14 = &vromimage[(map9_latch2_hi << 12) + 0x0400];
	PPU_bank_18 = &vromimage[(map9_latch2_hi << 12) + 0x0800];
	PPU_bank_1c = &vromimage[(map9_latch2_hi << 12) + 0x0c00];
    }
}

void map9_write(unsigned short address, unsigned char value)
{
    if (address < 0xa000) {
	/* Ignore it. It's not in spec. */
    } else if (address < 0xb000) {
	bank8_reg = value;
	sync_banks();
    } else if (address < 0xc000) {
	map9_latch1_lo = value;
	if (map9_latch1_state == 0) {
	    PPU_bank_00 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_04 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_08 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_0c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xd000) {
	map9_latch1_hi = value;
	if (map9_latch1_state == 1) {
	    PPU_bank_00 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_04 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_08 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_0c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xe000) {
	map9_latch2_lo = value;
	if (map9_latch2_state == 0) {
	    PPU_bank_10 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_14 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_18 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_1c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xf000) {
	map9_latch2_hi = value;
	if (map9_latch2_state == 1) {
	    PPU_bank_10 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_14 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_18 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_1c = &vromimage[(value << 12) + 0x0c00];
	}
    } else {
	if (value & 1) {
	    PPU_mirror_horizontal();
	} else {
	    PPU_mirror_vertical();
	}
    }
}

void map9_init(void)
{
    map9_latch1_state = 1;
    map9_latch2_state = 1;
    ppu_latchfunc = map9_latchfunc;
    bank8_reg = 0;
    bankA_reg = ((cart_rom_size - 1) << 1) - 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map9_iface = {map9_init, map9_write, dummy_hsync, dummy_vsync};


/* MAPPER A (10) Implementation */

unsigned char mapa_latch1_lo;
unsigned char mapa_latch1_hi;
unsigned char mapa_latch1_state;
unsigned char mapa_latch2_lo;
unsigned char mapa_latch2_hi;
unsigned char mapa_latch2_state;

void mapA_latchfunc(unsigned short address)
{
    if ((address & 0x3ff0) == 0x0fd0) {
	mapa_latch1_state = 0;
	PPU_bank_00 = &vromimage[(mapa_latch1_lo << 12) + 0x0000];
	PPU_bank_04 = &vromimage[(mapa_latch1_lo << 12) + 0x0400];
	PPU_bank_08 = &vromimage[(mapa_latch1_lo << 12) + 0x0800];
	PPU_bank_0c = &vromimage[(mapa_latch1_lo << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x0fe0) {
	mapa_latch1_state = 1;
	PPU_bank_00 = &vromimage[(mapa_latch1_hi << 12) + 0x0000];
	PPU_bank_04 = &vromimage[(mapa_latch1_hi << 12) + 0x0400];
	PPU_bank_08 = &vromimage[(mapa_latch1_hi << 12) + 0x0800];
	PPU_bank_0c = &vromimage[(mapa_latch1_hi << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x1fd0) {
	mapa_latch2_state = 0;
	PPU_bank_10 = &vromimage[(mapa_latch2_lo << 12) + 0x0000];
	PPU_bank_14 = &vromimage[(mapa_latch2_lo << 12) + 0x0400];
	PPU_bank_18 = &vromimage[(mapa_latch2_lo << 12) + 0x0800];
	PPU_bank_1c = &vromimage[(mapa_latch2_lo << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x1fe0) {
	mapa_latch2_state = 1;
	PPU_bank_10 = &vromimage[(mapa_latch2_hi << 12) + 0x0000];
	PPU_bank_14 = &vromimage[(mapa_latch2_hi << 12) + 0x0400];
	PPU_bank_18 = &vromimage[(mapa_latch2_hi << 12) + 0x0800];
	PPU_bank_1c = &vromimage[(mapa_latch2_hi << 12) + 0x0c00];
    }
}

void mapA_write(unsigned short address, unsigned char value)
{
    if (address < 0xa000) {
	/* Ignore it. It's not in spec. */
    } else if (address < 0xb000) {
	bank8_reg = value << 1;
	bankA_reg = (value << 1) + 1;
	sync_banks();
    } else if (address < 0xc000) {
	mapa_latch1_lo = value;
	if (mapa_latch1_state == 0) {
	    PPU_bank_00 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_04 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_08 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_0c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xd000) {
	mapa_latch1_hi = value;
	if (mapa_latch1_state == 1) {
	    PPU_bank_00 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_04 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_08 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_0c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xe000) {
	mapa_latch2_lo = value;
	if (mapa_latch2_state == 0) {
	    PPU_bank_10 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_14 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_18 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_1c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xf000) {
	mapa_latch2_hi = value;
	if (mapa_latch2_state == 1) {
	    PPU_bank_10 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_14 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_18 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_1c = &vromimage[(value << 12) + 0x0c00];
	}
    } else {
	if (value & 1) {
	    PPU_mirror_horizontal();
	} else {
	    PPU_mirror_vertical();
	}
    }
}

void mapA_init(void)
{
    mapa_latch1_state = 1;
    mapa_latch2_state = 1;
    ppu_latchfunc = mapA_latchfunc;
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t mapA_iface = {mapA_init, mapA_write, dummy_hsync, dummy_vsync};


/* MAPPER B (11) Implementation */

void mapB_write(unsigned short address, unsigned char value)
{
    unsigned char vromptr;
    unsigned char romptr;

    vromptr = (value >> 4) & 0x0f;
    if ((vromptr) > (cart_vrom_size >> 1)) {
	int i;

	i = 0x7;
	while ((vromptr &= i) >= cart_vrom_size) {
	    i >>= 1;
	}
    }

    romptr = value & 0x0f;
    if ((romptr) > (cart_rom_size >> 1)) {
	int i;

	i = 0x7;
	while ((romptr &= i) >= cart_rom_size) {
	    i >>= 1;
	}
    }
    
    PPU_bank_00 = &vromimage[(vromptr << 13)+0x0000];
    PPU_bank_04 = &vromimage[(vromptr << 13)+0x0400];
    PPU_bank_08 = &vromimage[(vromptr << 13)+0x0800];
    PPU_bank_0c = &vromimage[(vromptr << 13)+0x0c00];
    PPU_bank_10 = &vromimage[(vromptr << 13)+0x1000];
    PPU_bank_14 = &vromimage[(vromptr << 13)+0x1400];
    PPU_bank_18 = &vromimage[(vromptr << 13)+0x1800];
    PPU_bank_1c = &vromimage[(vromptr << 13)+0x1c00];
    bank8_reg = (romptr << 2);
    bankA_reg = (romptr << 2) + 1;
    bankC_reg = (romptr << 2) + 2;
    bankE_reg = (romptr << 2) + 3;
    sync_banks();
}

void mapB_init(void)
{
    ppu_latchfunc = NULL;
    mapB_write(0, 0);
    sync_banks();
}

mapper_iface_t mapB_iface = {mapB_init, mapB_write, dummy_hsync, dummy_vsync};


/*
 * $Log: mappers.c,v $
 * Revision 1.36  1998/12/20 03:43:13  nyef
 * updated mapper_support[] table.
 *
 * Revision 1.35  1998/12/20 03:31:32  nyef
 * added second latch to mapper 9.
 * mapper 9 now works just about perfectly.
 *
 * Revision 1.34  1998/12/15 04:41:27  nyef
 * added support for 512k and 1024k mapper 1 carts.
 *
 * Revision 1.33  1998/12/15 03:50:14  nyef
 * fixed bank mask bug in mapper 1.
 *
 * Revision 1.32  1998/12/15 03:43:20  nyef
 * reorganized the mapper 1 implementation to be much cleaner looking.
 *
 * Revision 1.31  1998/12/15 01:50:24  nyef
 * changed mapper hsync and vsync routines to conform to new spec.
 *
 * Revision 1.30  1998/11/28 03:59:31  nyef
 * changed map4_hsync() to postdecrement rather than predecrement the irq
 * count. Crystalis looks slightly better.
 *
 * Revision 1.29  1998/11/16 23:25:53  nyef
 * fixed mistake in the mapper 10 latchfunc.
 *
 * Revision 1.28  1998/11/05 03:07:40  nyef
 * added default VRom mapping to mapper 1.
 *
 * Revision 1.27  1998/11/03 01:41:24  nyef
 * added preliminary support for mapper 10.
 *
 * Revision 1.26  1998/11/01 22:09:40  nyef
 * commented out mapper 4 mirror control deb_printf.
 *
 * Revision 1.25  1998/10/04 14:14:16  nyef
 * fixed out-range bank compensation on mappers 11, 7, 3, 2 and 1.
 *
 * Revision 1.24  1998/10/04 14:10:52  nyef
 * fixed out-range rom bank compensation on mapper 4.
 * this will probably need to be propagated to other mappers.
 *
 * Revision 1.23  1998/10/03 23:18:36  nyef
 * added out-range rom bank compensation to mapper 4.
 *
 * Revision 1.22  1998/10/03 23:10:58  nyef
 * changed mapper 1 to only have 1 shift register.
 * this may improve some games.
 *
 * Revision 1.21  1998/10/03 13:46:18  nyef
 * added more code to mapper 11. (still doesn't work)
 *
 * Revision 1.20  1998/10/01 02:43:16  nyef
 * added preliminary support for mapper B (11).
 *
 * Revision 1.19  1998/10/01 02:35:17  nyef
 * added support for mapper 9.
 *
 * Revision 1.18  1998/09/01 01:00:01  nyef
 * fixed mapper 0. (added vrom stuff. oops.)
 *
 * Revision 1.17  1998/08/30 17:07:10  nyef
 * changed to use the new ppu_flags system instead of PPU_use_vrom.
 *
 * Revision 1.16  1998/08/24 01:57:53  nyef
 * fixed bug with 8k vrom pages on mapper 1.
 *
 * Revision 1.15  1998/08/24 01:24:19  nyef
 * more work on mapper 4.
 *
 * Revision 1.14  1998/08/24 00:29:06  nyef
 * more changes to mapper 4.
 *
 * Revision 1.13  1998/08/23 22:42:46  nyef
 * added hooks for calling mappers on vsync and hsync.
 * added preliminary irq support to mapper 4.
 *
 * Revision 1.12  1998/08/23 18:59:16  nyef
 * more changes to mapper 1. effectiveness unknown.
 *
 * Revision 1.11  1998/08/22 19:33:57  nyef
 * changes to mapper 4. effectiveness unknown.
 *
 * Revision 1.10  1998/08/22 18:41:37  nyef
 * more changes to mapper 1.
 *
 * Revision 1.9  1998/08/22 17:53:22  nyef
 * fixed more bugs in the mapper 1 emulation.
 *
 * Revision 1.8  1998/08/22 16:37:59  nyef
 * added support for mapper 1 roms that don't use VROM.
 *
 * Revision 1.7  1998/08/22 16:33:12  nyef
 * added preliminary VROM support to mapper 4.
 *
 * Revision 1.6  1998/08/04 21:56:31  nyef
 * changed the bit masks on the vrom selectors for mapper 1.
 *
 * Revision 1.5  1998/08/04 02:03:04  nyef
 * added an initialization message to set_mapper().
 *
 * Revision 1.4  1998/08/02 13:46:03  nyef
 * changed every printf to deb_printf for debug console.
 *
 * Revision 1.3  1998/07/31 02:28:00  nyef
 * upgraded the mapper 1 emulation somewhat. still incomplete, but better.
 *
 * Revision 1.2  1998/07/18 21:55:52  nyef
 * added preliminary mapper 7 support
 *
 * Revision 1.1  1998/07/11 22:18:27  nyef
 * Initial revision
 *
 */
