/*
 * nes_blit.c
 *
 * nes blit compiler
 */

/* $Id: nes_blit.c,v 1.7 1998/10/09 00:53:31 nyef Exp $ */

#include <stdlib.h>
#include "nes_blit.h"

/*
 * the minimum system requirements for this code are:
 *
 *    an assembly language programmer.
 *    a reference manual for a 386+ CPU.
 *    knowledge of function pointers in C.
 *    understanding of packed structures.
 *
 * if your system doesn't meet these requirements, then the code should still
 *   run, but you will be unable to understand or debug it.
 */

blitfunc_t nes_do_blit_80[0x02];
blitfunc_t nes_do_blit_c0[0x04];
blitfunc_t nes_do_blit_e0[0x08];
blitfunc_t nes_do_blit_f0[0x10];
blitfunc_t nes_do_blit_f8[0x20];
blitfunc_t nes_do_blit_fc[0x40];
blitfunc_t nes_do_blit_fe[0x80];
blitfunc_t nes_do_blit[0x100];
blitfunc_t nes_do_blit_7f[0x80];
blitfunc_t nes_do_blit_3f[0x40];
blitfunc_t nes_do_blit_1f[0x20];
blitfunc_t nes_do_blit_0f[0x10];
blitfunc_t nes_do_blit_07[0x08];
blitfunc_t nes_do_blit_03[0x04];
blitfunc_t nes_do_blit_01[0x02];

blitfunc_t nes_do_blit_rev_80[0x02];
blitfunc_t nes_do_blit_rev_c0[0x04];
blitfunc_t nes_do_blit_rev_e0[0x08];
blitfunc_t nes_do_blit_rev_f0[0x10];
blitfunc_t nes_do_blit_rev_f8[0x20];
blitfunc_t nes_do_blit_rev_fc[0x40];
blitfunc_t nes_do_blit_rev_fe[0x80];
blitfunc_t nes_do_blit_rev[0x100];
blitfunc_t nes_do_blit_rev_7f[0x80];
blitfunc_t nes_do_blit_rev_3f[0x40];
blitfunc_t nes_do_blit_rev_1f[0x20];
blitfunc_t nes_do_blit_rev_0f[0x10];
blitfunc_t nes_do_blit_rev_07[0x08];
blitfunc_t nes_do_blit_rev_03[0x04];
blitfunc_t nes_do_blit_rev_01[0x02];

/* bh: color 1 */
/* cl: color 2 */
/* ch: color 3 */
/* edi: cur_vbp */
/* bl: pattern */

/*
 * setup
 *
 * pushl %edi
 * pushl %ebx
 * pushl %ecx
 * movl 0x10(%esp), %edi
 * movw 0x14(%esp), %bx
 * movw 0x16(%esp), %cx
 * movb 0x18(%esp), %bl
 */

struct blit_hdr {
    unsigned char push1;  /* 0x57 */
    unsigned short push2; /* 0x5153 */
    unsigned short mov1a; /* 0x7c8b */
    unsigned short mov1b; /* 0x1024 */
    unsigned short mov2a; /* 0x8b66 */
    unsigned short mov2b; /* 0x245c */
    unsigned char mov2c;  /* 0x14 */
    unsigned short mov3a; /* 0x8b66 */
    unsigned short mov3b; /* 0x244c */
    unsigned char mov3c;  /* 0x16 */
    unsigned short mov4a; /* 0x5c8a */
    unsigned short mov4b; /* 0x1824 */
} __attribute__ ((packed));

struct blit_hdr blit_hdr = {0x57, 0x5153, 0x7c8b, 0x1024, 0x8b66, 0x245c, 0x14, 0x8b66, 0x244c, 0x16, 0x5c8a, 0x1824};

/*
 * for 0 bit
 *
 * rcl $1, %bl
 * jnc 0f
 * movb %bh, (%edi)
 * 0:
 * incl %edi
 */

struct blit_0 {
    unsigned short rcl; /* 0xd3d0 */
    unsigned short jnc; /* 0x0273 */
    unsigned short mov; /* 0x3f88 */
    unsigned char inc;  /* 0x47 */
} __attribute__ ((packed));

struct blit_0 blit_0 = {0xd3d0, 0x0273, 0x3f88, 0x47};
struct blit_0 blit_0_rev = {0xdbd0, 0x0273, 0x3f88, 0x47};

/*
 * for 1 bit
 *
 * rcl $1, %bl
 * jnc 0f
 * movb %ch, (%edi)
 * jmp 1f
 * 0:
 * movb %cl, (%edi)
 * 1:
 * incl %edi
 */

struct blit_1 {
    unsigned short rcl;  /* 0xd3d0 */
    unsigned short jnc;  /* 0x0473 */
    unsigned short mov1; /* 0x2f88 */
    unsigned short jmp;  /* 0x02eb */
    unsigned short mov2; /* 0x0f88 */
    unsigned char inc;   /* 0x47 */
} __attribute__ ((packed));

struct blit_1 blit_1 = {0xd3d0, 0x0473, 0x2f88, 0x02eb, 0x0f88, 0x47};
struct blit_1 blit_1_rev = {0xdbd0, 0x0473, 0x2f88, 0x02eb, 0x0f88, 0x47};

/*
 * cleanup
 *
 * popl %ecx
 * popl %ebx
 * popl %edi
 * ret
 */

struct blit_cleanup {
    unsigned short pop;     /* 0x5b59 */
    unsigned short cleanup; /* 0xc35f */
} __attribute__ ((packed));

struct blit_cleanup blit_cleanup = {0x5b59, 0xc35f};

#define BFL_NOCLIP    0
#define BFL_RIGHTCLIP 1
#define BFL_LEFTCLIP  2
#define BFL_CLIPMASK  3
#define BFL_REVERSED  4

/* if leftside is nonzero, all blitters are aligned left. */
void initialize_blitstream(blitfunc_t *table, int width, int flags)
{
    int i;
    int j;
    unsigned char *code_ptr;

    /*
     * there are 256 blitters. hence 256 headers and cleanup routines.
     * for each pair of blitters x and y where (y = 255 - x), there are
     * eight 1 bits and eight 0 bits. taking this data, we come up with
     * the following calculation for how much memory we to allocate.
     *
     * this formula has since been modified for arbitrary length blitters and
     * the overhead of clipping the image.
     */
#define SIZECALC(clipside) (((1 << width) * (sizeof(blit_hdr) + sizeof(blit_cleanup))) + ((1 << (width - 1)) * width * (sizeof(blit_0) + sizeof(blit_1))) + ((flags & clipside)? (3 * (1 << width)): 0));
    if ((flags & BFL_REVERSED) == 0) {
        i = SIZECALC(BFL_LEFTCLIP);
    } else {
	i = SIZECALC(BFL_RIGHTCLIP);
    }
#undef SIZECALC

    code_ptr = malloc(i);
    if (code_ptr == NULL) {
	deb_printf("Unnable to allocate %d bytes for compiled blitstream.\n", i);
        exit(1);
#if 0 /* for debugging only. */
    } else {
	deb_printf("Allocated %d bytes for compiled blitstream.\n", i);
#endif
    }

    for (i = 0; i < (1 << width); i++) {
	table[i] = (blitfunc_t)code_ptr;

	memcpy(code_ptr, &blit_hdr, sizeof(blit_hdr));
	code_ptr += sizeof(blit_hdr);

	if ((flags & BFL_CLIPMASK) != BFL_NOCLIP) {
	    if ((flags & BFL_LEFTCLIP) && ((flags & BFL_REVERSED) == 0)) {
		*((unsigned short *)code_ptr)++ = 0xd3c0;
		*code_ptr++ = (8 - width) & 0xff;
	    } else if (((flags & BFL_LEFTCLIP) == 0) && (flags & BFL_REVERSED)) {
		*((unsigned short *)code_ptr)++ = 0xdbc0;
		*code_ptr++ = (8 - width) & 0xff;
	    }
	}
	
	for (j = 0; j < width; j++) {
	    if ((flags & BFL_REVERSED) == 0) {
		if (i & (1 << ((width - 1) - j))) {
		    memcpy(code_ptr, &blit_1, sizeof(blit_1));
		    code_ptr += sizeof(blit_1);
		} else {
		    memcpy(code_ptr, &blit_0, sizeof(blit_0));
		    code_ptr += sizeof(blit_0);
		}
	    } else {
		if (i & (1 << j)) {
		    memcpy(code_ptr, &blit_1_rev, sizeof(blit_1));
		    code_ptr += sizeof(blit_1);
		} else {
		    memcpy(code_ptr, &blit_0_rev, sizeof(blit_0));
		    code_ptr += sizeof(blit_0);
		}
	    }
	}

	memcpy(code_ptr, &blit_cleanup, sizeof(blit_cleanup));
	code_ptr += sizeof(blit_cleanup);
    }
}


void blit_init(void)
{
    initialize_blitstream(nes_do_blit_80, 1, BFL_RIGHTCLIP);
    initialize_blitstream(nes_do_blit_c0, 2, BFL_RIGHTCLIP);
    initialize_blitstream(nes_do_blit_e0, 3, BFL_RIGHTCLIP);
    initialize_blitstream(nes_do_blit_f0, 4, BFL_RIGHTCLIP);
    initialize_blitstream(nes_do_blit_f8, 5, BFL_RIGHTCLIP);
    initialize_blitstream(nes_do_blit_fc, 6, BFL_RIGHTCLIP);
    initialize_blitstream(nes_do_blit_fe, 7, BFL_RIGHTCLIP);
    initialize_blitstream(nes_do_blit, 8, BFL_NOCLIP);
    initialize_blitstream(nes_do_blit_7f, 7, BFL_LEFTCLIP);
    initialize_blitstream(nes_do_blit_3f, 6, BFL_LEFTCLIP);
    initialize_blitstream(nes_do_blit_1f, 5, BFL_LEFTCLIP);
    initialize_blitstream(nes_do_blit_0f, 4, BFL_LEFTCLIP);
    initialize_blitstream(nes_do_blit_07, 3, BFL_LEFTCLIP);
    initialize_blitstream(nes_do_blit_03, 2, BFL_LEFTCLIP);
    initialize_blitstream(nes_do_blit_01, 1, BFL_LEFTCLIP);

    initialize_blitstream(nes_do_blit_rev_80, 1, BFL_RIGHTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_c0, 2, BFL_RIGHTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_e0, 3, BFL_RIGHTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_f0, 4, BFL_RIGHTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_f8, 5, BFL_RIGHTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_fc, 6, BFL_RIGHTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_fe, 7, BFL_RIGHTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev, 8, BFL_NOCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_7f, 7, BFL_LEFTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_3f, 6, BFL_LEFTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_1f, 5, BFL_LEFTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_0f, 4, BFL_LEFTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_07, 3, BFL_LEFTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_03, 2, BFL_LEFTCLIP | BFL_REVERSED);
    initialize_blitstream(nes_do_blit_rev_01, 1, BFL_LEFTCLIP | BFL_REVERSED);
}

/*
 * $Log: nes_blit.c,v $
 * Revision 1.7  1998/10/09 00:53:31  nyef
 * fixed debug output.
 *
 * Revision 1.6  1998/09/26 13:24:08  nyef
 * fixed memory size calculation bug for handling reversed clipped blitters.
 * also fixed blitclip code for same case.
 *
 * Revision 1.5  1998/08/15 15:20:42  nyef
 * added blitters for reversed operation (sprites).
 *
 * Revision 1.4  1998/08/15 04:58:49  nyef
 * added blitters for clipped left-side operation.
 *
 * Revision 1.3  1998/08/15 04:14:50  nyef
 * added blitters for clipped right-side operation.
 *
 * Revision 1.2  1998/07/27 02:36:24  nyef
 * changed initialization to use 255 fewer calls to malloc(3) and folded the
 * two helper routines into the main code.
 *
 * Revision 1.1  1998/07/25 16:56:41  nyef
 * Initial revision
 *
 */
