/*
 * opcz80.h
 *
 * z80 opcode emulation macros
 */

/* $Id: opcz80.h,v 1.1 1999/01/01 20:43:53 nyef Exp $ */

#ifndef __OPCZ80_H__
#define __OPCZ80_H__

#include "machdep.h"

#define reg_F       0x0(%ebp)
#define reg_A       0x1(%ebp)
#define reg_BC      0x2(%ebp)
#define reg_C       0x2(%ebp)
#define reg_B       0x3(%ebp)
#define reg_DE      0x4(%ebp)
#define reg_E       0x4(%ebp)
#define reg_D       0x5(%ebp)
#define reg_HL      0x6(%ebp)
#define reg_L       0x6(%ebp)
#define reg_H       0x7(%ebp)

/* alternate register set goes here */

#define reg_I       0x10(%ebp)
#define reg_R       0x11(%ebp)
#define reg_IX      0x12(%ebp)
#define reg_IY      0x14(%ebp)
#define reg_SP      0x16(%ebp)
#define reg_PC      0x18(%ebp)

#define readmap     0x1a(%ebp)
#define writemap    0x1e(%ebp)
#define cycles_left 0x22(%ebp)
#define mmu_shift   0x26(%ebp)
#define mmu_mask    0x2a(%ebp)
#define user        0x2e(%ebp)
#define dispatch    0x32(%ebp)
#define decodemap   0x36(%ebp)

    CYC_ADDMODE = 0
    CYC_OP = 1
#define Continue subl $(CYC_ADDMODE+CYC_OP), cycles_left; jmp *dispatch; CYC_ADDMODE = 0; CYC_OP = 1

/* emulation register map */
/* %edi = effective address */
/* %esi = memaccess scratch */ /* used only in the GetInd(X/Y) macros */
/* %edx = PC */
/* %ch = F */
/* %ax = opcode scratch */

#define MemRead pushl %edx; pushl %ecx; pushl %eax; pushl %edi; pushl user; movl mmu_shift, %ecx; movl mmu_mask, %edx; shrl %cl, %edi; andl %edx, %edi; movl readmap, %esi; call (%esi,%edi,4); movb %al, 8(%esp); popl %edi; popl %edi; popl %eax; popl %ecx; popl %edx
#define MemReadW incw %di; MemRead; decw %di; shlw $8, %ax; MemRead

#define MemWrite pushl %edx; pushl %ecx; pushl %eax; pushl %edi; pushl user; movl mmu_shift, %ecx; movl mmu_mask, %ebx; shrl %cl, %edi; andl %ebx, %edi; movl writemap, %esi; call (%esi,%edi,4); popl %edi; popl %edi; popl %eax; popl %ecx; popl %edx

#define GetFromPC_B movl %edx, %edi; incw %dx
#define GetFromPC_W movl %edx, %edi; incw %dx; incw %dx

#define push(x) xorl %edi, %edi; decw reg_SP; movw reg_SP, %di; movb x, %al; MemWrite
#define pop(x) xorl %edi, %edi; movw reg_SP, %di; incw reg_SP; MemRead; movb %al, x

#endif /* __OPCZ80_H__ */

/*
 * $Log: opcz80.h,v $
 * Revision 1.1  1999/01/01 20:43:53  nyef
 * Initial revision
 *
 */
