/*
 * pce.c
 *
 * contains PC-Engine specific code and data.
 */

/* $Id: pce.c,v 1.14 1999/01/17 02:12:43 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>

#include "emu6280.h"
#include "tool.h"
#include "system.h"
#include "debug.h"
#include "cal.h"
#include "ui.h"
#include "pce_vdp.h"

cal_cpu pce_cpu;
pce_vdp pce_video;

#define PCE_RAMSIZE 0x8000

unsigned char *pce_mmap[256];

unsigned char pce_read_mem(cal_cpu cpu, unsigned short addr)
{
    if (cpu->data.d_emu6280->mpr[(addr >> 13) & 7] == 0xff) {
	switch (addr & 0x1c00) {
	case 0x0000:
	    switch (addr & 3) {
	    case 0:
		return pce_video->readvdp0l(pce_video);
		break;
	    case 1:
		return pce_video->readvdp0h(pce_video);
		break;
	    case 2:
		return pce_video->readvdp1l(pce_video);
		break;
	    case 3:
		return pce_video->readvdp1h(pce_video);
		break;
	    }
	    break;
	default:
/* 	    deb_printf("pce: I/O read 0x%04hx, returning 0x00.\n", addr & 0x1fff); */
	}
	return 0x00;
    } else {
	return pce_mmap[cpu->data.d_emu6280->mpr[(addr >> 13) & 7]][addr & 0x1fff];
    }
}

void pce_write_mem(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    if (cpu->data.d_emu6280->mpr[(addr >> 13) & 7] == 0xff) {
	switch (addr & 0x1c00) {
	case 0x0000:
	    switch (addr & 3) {
	    case 0:
		pce_video->writevdp0l(pce_video, data);
		break;
	    case 1:
		pce_video->writevdp0h(pce_video, data);
		break;
	    case 2:
		pce_video->writevdp1l(pce_video, data);
		break;
	    case 3:
		pce_video->writevdp1h(pce_video, data);
		break;
	    }
	    break;
	default:
/* 	    deb_printf("pce: I/O write 0x%04hx = 0x%02x.\n", addr & 0x1fff, data); */
	}
    } else if (cpu->data.d_emu6280->mpr[(addr >> 13) & 7] == 0xf7) {
	/* Savegame RAM */
    } else if (cpu->data.d_emu6280->mpr[(addr >> 13) & 7] > 0xf7) {
	/* Normal RAM */
    } else {
	/* Ignore it, it's in the ROM area */
    }
}

memread_t pce_memread[1] = {pce_read_mem,};
memwrite_t pce_memwrite[1] = {pce_write_mem,};

void pce_timeslice(cal_cpu cpu)
{
    int i;

    for (i = 243; i && (system_flags == F_NONE); i--) {
	cpu->runfor(cpu, 115);
	cpu->run(cpu);
	if (pce_video->periodic(pce_video)) {
/* 	    deb_printf("pce_video->periodic() requests interrupt.\n"); */
	    cpu->irq(cpu);
	}
	if (system_flags & F_UNIMPL) {
/* 	    deb_printf("Unimplimented Instruction.\n"); */
	    deb_printf("Unimplimented Instruction 0x%02x.\n", pce_read_mem(cpu, cpu->data.d_emu6280->reg_PC));
	}
    }
}

void pce_init_mmap(void)
{
    int i;
    int offset;

    offset = 0;
    
    if (romsize & 0x200) {
	printf("pce_init_mmap(): compensating for 512-byte header.\n");
	offset = 0x200;
    }
    
    for (i = 0; i < 0xf7; i++) {
	pce_mmap[i] = romimage +
	    ((((i * 0x2000) + offset) < romsize)? ((i * 0x2000) + offset): 0);
    }

    pce_mmap[0xf7] = NULL;

    for (i = 0xf8; i < 0xfc; i++) {
	pce_mmap[i] = ramblock + ((i - 0xf8) + 0x2000);
    }

    for (i = 0xfc; i < 0x100; i++) {
	pce_mmap[i] = NULL;
    }
}

void pce_run(int startup_deb)
{
    ramblock = malloc(PCE_RAMSIZE);

    pce_init_mmap();
    pce_cpu = cal_create(CPUT_EMU6280);

    pce_video = pce_vdp_create();
    
    if ((romimage != NULL) && (ramblock != NULL) && (pce_cpu) && (pce_video)) {
	pce_cpu->setzpage(pce_cpu, ramblock);
	pce_cpu->setmmu(pce_cpu, 0, 0, pce_memread, pce_memwrite);
	pce_cpu->reset(pce_cpu);
	
	set_timeslice(&pce_timeslice, pce_cpu);
    } else {
	printf("out of memory\n");
    }
}

/*
 * $Log: pce.c,v $
 * Revision 1.14  1999/01/17 02:12:43  nyef
 * added support for the new pce vdp
 *
 * Revision 1.13  1999/01/13 03:55:36  nyef
 * fixed to use the cal, the timeslicer, and the new memory access spec.
 *
 * Revision 1.12  1998/07/27 00:57:40  nyef
 * added compensation for 512-byte header in pce_init_mmap()
 *
 * Revision 1.11  1998/07/26 23:11:56  nyef
 * changed read/write table allocation to allocate extra room for zpage
 *
 * Revision 1.10  1998/07/14 10:25:21  nyef
 * added temporary assignments to cycles_left in pce_trace and pce_freerun
 *
 * Revision 1.9  1998/07/14 01:52:38  nyef
 * added initializer for disassemble()
 *
 * Revision 1.8  1998/07/14 01:37:21  nyef
 * added trace function for debugger
 *
 * Revision 1.7  1998/07/12 23:22:00  nyef
 * fixed two stupid mistakes in the mmap initializer
 *
 * Revision 1.6  1998/07/12 23:21:03  nyef
 * added initializer for memory mapping
 *
 * Revision 1.5  1998/07/12 23:03:07  nyef
 * fixed mem_read_byte()
 *
 * Revision 1.4  1998/07/12 22:36:05  nyef
 * set ram size to 32k (this figure is enough to cover the SuperGrafx)
 *
 * Revision 1.3  1998/07/12 21:50:00  nyef
 * finished fixing display_vectors()
 *
 * Revision 1.2  1998/07/12 21:24:57  nyef
 * started fixing display_vectors()
 * added freerun()
 * made regs static
 *
 * Revision 1.1  1998/07/11 22:17:59  nyef
 * Initial revision
 *
 */
