/*
 * pce_vdp.c
 *
 * HuC6270/HuC6260 emulation.
 */

/* $Id: pce_vdp.c,v 1.9 1999/02/01 04:00:30 nyef Exp $ */

#include <stdlib.h>
#include "pce_vdp.h"
#include "ui.h"
#include "video.h"
#include "blitters.h"

#define PCE_VDP_RAMSIZE 0x10000

int pce_palbase_red[64] = {
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
};

int pce_palbase_blue[64] = {
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
};

int pce_palbase_green[64] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55,
    0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa,
    0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
};

unsigned char pce_vce_read(pce_vdp vdp, unsigned short addr)
{
    unsigned char retval;
    
    if ((addr & 6) == 4) {
	if (addr & 1) {
	    retval = vdp->palette_l[vdp->palette_addr];
	    vdp->palette_addr += 1;
	    vdp->palette_addr &= 0x1ff;
	    return retval;
	} else {
	    return vdp->palette_l[vdp->palette_addr];
	}
    }
    return 0x00;
}

#define PCE_COLOR_XLAT(x) vid_pre_xlat[((vdp->palette_l[x] & 0x6) >> 1) | ((vdp->palette_l[x] & 0x30) >> 2) | ((vdp->palette_l[x] & 0x80) >> 3) | ((vdp->palette_h[x] & 1) << 5)]
void pce_vce_write(pce_vdp vdp, unsigned short addr, unsigned char data)
{
    switch (addr & 7) {
    case 0:
	/* FIXME: add control register */
	break;
    case 2:
	vdp->palette_addr &= 0x100;
	vdp->palette_addr |= data;
	break;
    case 3:
	if (data & 1) {
	    vdp->palette_addr |= 0x100;
	} else {
	    vdp->palette_addr &= 0xff;
	}
	break;
    case 4:
	vdp->palette_l[vdp->palette_addr] = data;
	vdp->palette_xlat[vdp->palette_addr] = PCE_COLOR_XLAT(vdp->palette_addr);
	break;
    case 5:
	vdp->palette_h[vdp->palette_addr] = data & 1;
	vdp->palette_xlat[vdp->palette_addr] = PCE_COLOR_XLAT(vdp->palette_addr);
	vdp->palette_addr += 1;
	vdp->palette_addr &= 0x1ff;
	break;
    }
}

unsigned char pce_vdp_read(pce_vdp vdp, unsigned short addr)
{
    unsigned char retval;
    
    switch (addr & 3) {
    case 0:
	deb_printf("pce_vdp: statread.\n");
	retval = vdp->status;
	vdp->status &= ~0x20; /* clear vblank bit */ /* FIXME: may be wrong */
	return retval;
    case 1:
	return 0x00;
    case 2:
	if (vdp->cur_reg != 2) {
	    deb_printf("pce_vdp: r1l cur_reg %d.\n", vdp->cur_reg);
	} else {
	    deb_printf("pce_vdp: memread low.\n");
	    /* FIXME: Don't just do something, stand there! */
	}
	return 0x00;
    case 3:
	if (vdp->cur_reg != 2) {
	    deb_printf("pce_vdp: r1h cur_reg %d.\n", vdp->cur_reg);
	} else {
	    deb_printf("pce_vdp: memread high.\n");
	    /* FIXME: Don't just do something, stand there! */
	}
	return 0x00;
    }
    return 0x00;
}

void pce_vdp_write(pce_vdp vdp, unsigned short addr, unsigned char data)
{
    switch (addr & 3) {
    case 0:
	vdp->cur_reg = data & 0x1f;
	break;
    case 1:
	break;
    case 2:
	switch (vdp->cur_reg) {
	case 0:
	    vdp->mem_write_addr &= 0xff00;
	    vdp->mem_write_addr |= data;
	    break;
	case 1:
	    vdp->mem_read_addr &= 0xff00;
	    vdp->mem_read_addr |= data;
	    break;
	case 2:
	    vdp->memory[(vdp->mem_write_addr << 1) + 0] = data;
	    break;
	case 5: /* Control */
	    vdp->regs_l[5] = data;
	    break;
	case 6: /* Raster counter */
	    break;
	case 7: /* BG X-Scroll */
	    break;
	case 8: /* BG Y-Scroll */
	    break;
	case 9: /* Memory width? */
	    break;
	case 10: /* Horizontal Sync? */
	    break;
	case 11: /* Horizontal Display? */
	    break;
	case 12: /* Vertical Sync? */
	    break;
	case 13: /* Vertical Display? */
	    break;
	case 14: /* Vertical Display End? */
	    break;
	case 15: /* DMA Control */
	    break;
	case 19: /* Sprite attributes? */
	    break;
	default:
	    deb_printf("pce_vdp: regwrite %d l = 0x%02x.\n", vdp->cur_reg, data);
	}
	break;
    case 3:
	switch (vdp->cur_reg) {
	case 0:
	    vdp->mem_write_addr &= 0x00ff;
	    vdp->mem_write_addr |= (data << 8);
	    break;
	case 1:
	    vdp->mem_read_addr &= 0x00ff;
	    vdp->mem_read_addr |= (data << 8);
	    break;
	case 2:
	    vdp->memory[(vdp->mem_write_addr << 1) + 1] = data;
	    vdp->mem_write_addr += (1 << ((vdp->regs_h[5] & 0x18)? ((vdp->regs_h[5] & 0x18) >> 3) + 4: 0)); /* FIXME: is this right? */
	    break;
	case 5: /* Control */
	    vdp->regs_h[5] = data;
	    break;
	case 6: /* Raster counter */
	    break;
	case 7: /* BG X-Scroll */
	    break;
	case 8: /* BG Y-Scroll */
	    break;
	case 9: /* Memory width? */
	    break;
	case 10: /* Horizontal Sync? */
	    break;
	case 11: /* Horizontal Display? */
	    break;
	case 12: /* Vertical Sync? */
	    break;
	case 13: /* Vertical Display? */
	    break;
	case 14: /* Vertical Display End? */
	    break;
	case 15: /* DMA Control */
	    break;
	case 19: /* Sprite attributes? */
	    break;
	default:
	    deb_printf("pce_vdp: regwrite %d h = 0x%02x.\n", vdp->cur_reg, data);
	}
	break;
    }
}

void pce_vdp_render_background(pce_vdp vdp, unsigned char *cur_vbp)
{
    unsigned char *pattern_table;
    unsigned short *cur_pos;
    unsigned short cur_tile;
    unsigned char *cur_pattern;
    int cur_pattern_line;
    int i;

    pattern_table = vdp->memory;
    cur_pattern_line = vdp->cur_scanline & 7;
    cur_pos = &((unsigned short *)vdp->memory)[(vdp->cur_scanline & 0xf8) << 3];
    pattern_table += cur_pattern_line << 1;

    for (i = 0; i < 32; i++) {
	cur_tile = *cur_pos++;
	cur_pattern = &pattern_table[(cur_tile & 0xfff) << 5];
	
	blit_4_8(cur_pattern[17], cur_pattern[16], cur_pattern[1], cur_pattern[0], 8, 0, &vdp->palette_xlat[(cur_tile >> 8) & 0xf0], cur_vbp);
	cur_vbp += 8;
    }
}

void pce_vdp_render_line(pce_vdp vdp)
{
    unsigned char *cur_vbp;

    cur_vbp = video_get_vbp(vdp->cur_scanline);

    if (vdp->regs_l[5] & 0x80) {
	pce_vdp_render_background(vdp, cur_vbp);
    } else {
	/* FIXME: floodfill with black */
    }
}

int pce_vdp_periodic(pce_vdp vdp)
{
    if (vdp->cur_scanline < 256) {
	pce_vdp_render_line(vdp);
    }
    if (vdp->cur_scanline == 256) {
	video_events();
	video_display_buffer();
	vdp->status |= 0x20; /* FIXME: I'm not sure about this. */
    }
    if (vdp->cur_scanline == 261) {
	vdp->cur_scanline = 0;
    } else {
	vdp->cur_scanline++;
    }
    return ((vdp->regs_l[5] & 0x08) && (vdp->cur_scanline == 256));
#if 0
    return ((vdp->regs_l[5] & 0x08) && (vdp->status & 0x20)); /* FIXME: may be wrong. */
#endif
}

pce_vdp pce_vdp_create(void)
{
    pce_vdp retval;
    
    retval = calloc(1, sizeof(struct pce_vdp));
    if (retval) {
	retval->memory = calloc(1, PCE_VDP_RAMSIZE);
	if (retval->memory) {
	    retval->readvce = pce_vce_read;
	    retval->writevce = pce_vce_write;
	    retval->readvdp = pce_vdp_read;
	    retval->writevdp = pce_vdp_write;
	    retval->periodic = pce_vdp_periodic;
	    video_setpal(64, pce_palbase_red, pce_palbase_green, pce_palbase_blue);
	} else {
	    free(retval);
	    retval = NULL;
	}
    }
    
    if (!retval) {
	deb_printf("pce_vdp_create(): out of memory.\n");
    }
    
    return retval;
}

/*
 * $Log: pce_vdp.c,v $
 * Revision 1.9  1999/02/01 04:00:30  nyef
 * changed to only render lines < 256
 *
 * Revision 1.8  1999/01/31 21:19:50  nyef
 * removed some annoying debug output
 *
 * Revision 1.7  1999/01/31 20:33:26  nyef
 * fixed tile layer order in pce_vdp_render_background()
 *
 * Revision 1.6  1999/01/31 06:04:49  nyef
 * added very preliminary video renderer.
 * also added vce palette emulation (at 3 bpp less than the real system).
 *
 * Revision 1.5  1999/01/31 00:16:08  nyef
 * added preliminary implimentation of palette ram
 *
 * Revision 1.4  1999/01/30 23:26:11  nyef
 * added dummy routines for VCE I/O.
 *
 * Revision 1.3  1999/01/30 23:09:33  nyef
 * redesigned VDP I/O interface.
 *
 * Revision 1.2  1999/01/30 19:31:56  nyef
 * added some debug output for memory access.
 * changed pce_vdp_periodic() to interrupt by scanline rather
 * than status flag.
 *
 * Revision 1.1  1999/01/17 02:12:48  nyef
 * Initial revision
 *
 */
