/*
 * pce_vdp.c
 *
 * HuC6270/HuC6260 emulation.
 */

/* $Id: pce_vdp.c,v 1.1 1999/01/17 02:12:48 nyef Exp $ */

#include <stdlib.h>
#include "pce_vdp.h"
#include "ui.h"

#define PCE_VDP_RAMSIZE 0x10000

unsigned char pce_vdp_readvdp0l(pce_vdp vdp)
{
    unsigned char retval;
    
    deb_printf("pce_vdp: statread.\n");
    retval = vdp->status;
    vdp->status &= ~0x20; /* clear vblank bit */ /* FIXME: may be wrong */
    return retval;
}

unsigned char pce_vdp_readvdp0h(pce_vdp vdp)
{
    return 0x00; /* NOTE: this is correct. */
}

unsigned char pce_vdp_readvdp1l(pce_vdp vdp)
{
    if (vdp->cur_reg != 2) {
	deb_printf("pce_vdp: r1l cur_reg %d.\n", vdp->cur_reg);
    } else {
	/* FIXME: Don't just do something, stand there! */
    }
    return 0x00;
}

unsigned char pce_vdp_readvdp1h(pce_vdp vdp)
{
    if (vdp->cur_reg != 2) {
	deb_printf("pce_vdp: r1h cur_reg %d.\n", vdp->cur_reg);
    } else {
	/* FIXME: Don't just do something, stand there! */
    }
    return 0x00;
}

void pce_vdp_writevdp0l(pce_vdp vdp, unsigned char data)
{
    vdp->cur_reg = data & 0x1f;
}

void pce_vdp_writevdp0h(pce_vdp vdp, unsigned char data)
{
    /* NOTE: This is correct. */
}

void pce_vdp_writevdp1l(pce_vdp vdp, unsigned char data)
{
    switch (vdp->cur_reg) {
    case 0:
	vdp->mem_write_addr &= 0xff00;
	vdp->mem_write_addr |= data;
	break;
    case 1:
	vdp->mem_read_addr &= 0xff00;
	vdp->mem_read_addr |= data;
	break;
    case 2:
	vdp->memory[(vdp->mem_write_addr << 1) + 0] = data;
	break;
    case 5: /* Control */
	vdp->regs_l[5] = data;
	break;
    case 6: /* Raster counter */
	break;
    case 7: /* BG X-Scroll */
	break;
    case 8: /* BG Y-Scroll */
	break;
    case 9: /* Memory width? */
	break;
    case 10: /* Horizontal Sync? */
	break;
    case 11: /* Horizontal Display? */
	break;
    case 12: /* Vertical Sync? */
	break;
    case 13: /* Vertical Display? */
	break;
    case 14: /* Vertical Display End? */
	break;
    case 15: /* DMA Control */
	break;
    case 19: /* Sprite attributes? */
	break;
    default:
	deb_printf("pce_vdp: regwrite %d l = 0x%02x.\n", vdp->cur_reg, data);
    }
}

void pce_vdp_writevdp1h(pce_vdp vdp, unsigned char data)
{
    switch (vdp->cur_reg) {
    case 0:
	vdp->mem_write_addr &= 0x00ff;
	vdp->mem_write_addr |= (data << 8);
	break;
    case 1:
	vdp->mem_read_addr &= 0x00ff;
	vdp->mem_read_addr |= (data << 8);
	break;
    case 2:
	vdp->memory[(vdp->mem_write_addr << 1) + 1] = data;
	vdp->mem_write_addr += (1 << ((vdp->regs_h[5] & 0x18)? ((vdp->regs_h[5] & 0x18) >> 3) + 4: 0)); /* FIXME: is this right? */
	break;
    case 5: /* Control */
	vdp->regs_h[5] = data;
	break;
    case 6: /* Raster counter */
	break;
    case 7: /* BG X-Scroll */
	break;
    case 8: /* BG Y-Scroll */
	break;
    case 9: /* Memory width? */
	break;
    case 10: /* Horizontal Sync? */
	break;
    case 11: /* Horizontal Display? */
	break;
    case 12: /* Vertical Sync? */
	break;
    case 13: /* Vertical Display? */
	break;
    case 14: /* Vertical Display End? */
	break;
    case 15: /* DMA Control */
	break;
    case 19: /* Sprite attributes? */
	break;
    default:
	deb_printf("pce_vdp: regwrite %d h = 0x%02x.\n", vdp->cur_reg, data);
    }
}

int pce_vdp_periodic(pce_vdp vdp)
{
    if (vdp->cur_scanline == 256) {
	vdp->status |= 0x20; /* FIXME: I'm not sure about this. */
    }
    if (vdp->cur_scanline == 261) {
	vdp->cur_scanline = 0;
    } else {
	vdp->cur_scanline++;
    }
    return ((vdp->regs_l[5] & 0x08) && (vdp->status & 0x20)); /* FIXME: may be wrong. */
}

pce_vdp pce_vdp_create(void)
{
    pce_vdp retval;
    
    retval = calloc(1, sizeof(struct pce_vdp));
    if (retval) {
	retval->memory = calloc(1, PCE_VDP_RAMSIZE);
	if (retval->memory) {
	    retval->readvdp0l = pce_vdp_readvdp0l;
	    retval->readvdp0h = pce_vdp_readvdp0h;
	    retval->readvdp1l = pce_vdp_readvdp1l;
	    retval->readvdp1h = pce_vdp_readvdp1h;
	    retval->writevdp0l = pce_vdp_writevdp0l;
	    retval->writevdp0h = pce_vdp_writevdp0h;
	    retval->writevdp1l = pce_vdp_writevdp1l;
	    retval->writevdp1h = pce_vdp_writevdp1h;
	    retval->periodic = pce_vdp_periodic;
/* 	    video_setpal(64, sms_palbase_red, sms_palbase_green, sms_palbase_blue); */
	} else {
	    free(retval);
	    retval = NULL;
	}
    }
    
    if (!retval) {
	deb_printf("pce_vdp_create(): out of memory.\n");
    }
    
    return retval;
}

/*
 * $Log: pce_vdp.c,v $
 * Revision 1.1  1999/01/17 02:12:48  nyef
 * Initial revision
 *
 */
