/*
 * sms.c
 *
 * Sega Master System emulation.
 */

/* $Id: sms.c,v 1.12 1999/01/10 19:55:38 nyef Exp $ */

#include <stdlib.h>

#include "ui.h"
#include "system.h"
#include "cal.h"
#include "tool.h"
/* #include "emuz80.h" */
/* #include "mz80.h" */
#include "sms9918.h"

#define SMS_RAMSIZE 0xa000 /* 0x2000 ram + 0x8000 cart ram */

unsigned char *sms_memory;
cal_cpu sms_cpu;
sms9918 sms_vdp;

unsigned char *sms_rombase;
unsigned char *sms_bank0;
unsigned char *sms_bank1;
unsigned char *sms_bank2;
int sms_romsize;

unsigned char sms_pageregs[4];

void sms_pagesync(void)
{
    sms_bank0 = &sms_rombase[(sms_pageregs[1] << 14) & (sms_romsize - 1)];
    sms_bank1 = &sms_rombase[(sms_pageregs[2] << 14) & (sms_romsize - 1)];
    if (sms_pageregs[0] & 0x08) {
	sms_bank2 = &sms_memory[0x2000 + ((sms_pageregs[0] & 0x04) << 12)];
    } else {
	sms_bank2 = &sms_rombase[(sms_pageregs[3] << 14) & (sms_romsize - 1)];
    }
}

unsigned char sms_read_rom0(cal_cpu cpu, unsigned short addr)
{
    if (addr < 0x400) {
	return sms_rombase[addr];
    } else {
	return sms_bank0[addr];
    }
}

unsigned char sms_read_rom1(cal_cpu cpu, unsigned short addr)
{
    return sms_bank1[addr & 0x3fff];
}

unsigned char sms_read_rom2(cal_cpu cpu, unsigned short addr)
{
    return sms_bank2[addr & 0x3fff];
}

unsigned char sms_read_ram(cal_cpu cpu, unsigned short addr)
{
    if ((addr & 0x1fff) >= 0x1ffc) {
	return sms_pageregs[addr & 3];
    } else {
	return sms_memory[addr & 0x1fff];
    }
}

void sms_write_ignore(cal_cpu cpu, unsigned short addr, unsigned char data)
{
}

void sms_write_rom2(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    if (sms_pageregs[0] & 0x08) {
	sms_bank2[addr & 0x3fff] = data;
    }
}

void sms_write_ram(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    if (addr >= 0xfffc) {
	sms_pageregs[addr & 3] = data;
	sms_pagesync();
    } else {
	sms_memory[addr & 0x1fff] = data;
    }
}

memread_t sms_readtable[4] = {
    sms_read_rom0, sms_read_rom1, sms_read_rom2, sms_read_ram,
};

memwrite_t sms_writetable[4] = {
    sms_write_ignore, sms_write_ignore, sms_write_rom2, sms_write_ram,
};

unsigned char sms_io_read(cal_cpu cpu, unsigned short address)
{
    if ((address & 0xfe) == 0xbe) {
	if (address & 1) {
/* 	    deb_printf("sms: VDPstat read at PC 0x%04hx.\n", sms_cpu->data.d_maratz80->PC.W); */
	    return sms_vdp->readport1(sms_vdp);
	} else {
	    return sms_vdp->readport0(sms_vdp);
	}
    } else if (((address & 0xfe) == 0xdc) || ((address & 0xfe) == 0xc0)) {
	/* Dummy controller emulation */
	if (address & 1) {
	    return 0xff;
	} else {
	    extern unsigned char joy1_data;
	    return (((joy1_data & 0xf0) >> 4) | ((joy1_data & 3) << 4)) ^ 255;
	}
    } else if ((address & 0xfe) == 0x7e) {
	return sms_vdp->readscanline(sms_vdp);
    } else {
	deb_printf("sms: I/O read 0x%02x.\n", address & 0xff);
    }

    return 0;
}

void sms_io_write(cal_cpu cpu, unsigned short address, unsigned char data)
{
    if ((address & 0xfe) == 0xbe) {
	if (address & 1) {
	    sms_vdp->writeport1(sms_vdp, data);
	} else {
	    sms_vdp->writeport0(sms_vdp, data);
	}
    } else if ((address & 0xfe) == 0x7e) {
	/* I don't like seeing the sound I/O */
    } else if ((address & 0xfe) == 0xf0) {
	/* I don't like seeing the sound I/O */
    } else {
	deb_printf("sms: I/O write 0x%02x = 0x%02x.\n", address & 0xff, data);
    }
}

void sms_timeslice(cal_cpu cpu)
{
    int i;

    for (i = 262; i && (system_flags == F_NONE); i--) {
	cpu->runfor(cpu, 228); /* FIXME: this is .5 over what it should be. */
	cpu->run(cpu);
/* 	deb_printf("sms: PC 0x%04hx.\n", sms_cpu->data.d_maratz80->PC.W); */
	if (sms_vdp->periodic(sms_vdp)) {
	    cpu->irq(cpu);
	}
	if (system_flags & F_UNIMPL) {
	    deb_printf("Unimplimented Instruction.\n");
/* 	    deb_printf("PC: 0x%04hx, opcode: 0x%02x.\n", cpu->data.d_emuz80->reg_PC, sms_readtable[(cpu->data.d_emuz80->reg_PC >> 14) & 3](cpu, cpu->data.d_emuz80->reg_PC)); */
	}
    }
}

void sms_run(int startup_deb)
{
    sms_memory = calloc(1, SMS_RAMSIZE);
    
    if (sms_memory && romimage) {
	if (romsize & 0x200) {
	    sms_rombase = romimage + 0x200;
	    sms_romsize = romsize - 0x200;
	} else {
	    sms_rombase = romimage;
	    sms_romsize = romsize;
	}

	if (sms_rombase[0x7ffd] < 0x50) {
	    deb_printf("file is probably a Game Gear image.\n");
	} else {
	    deb_printf("file is probably a Master System image.\n");
	}
	sms_pageregs[0] = 0;
	sms_pageregs[1] = 0;
	sms_pageregs[2] = 1;
	sms_pageregs[3] = 2;
	sms_pagesync();
    
	sms_cpu = cal_create(CPUT_MARATZ80);
	sms_cpu->setmmu(sms_cpu, 14, 3, sms_readtable, sms_writetable);
	sms_cpu->setiou(sms_cpu, sms_io_read, sms_io_write);
	sms_cpu->reset(sms_cpu);

	sms_vdp = sms9918_create();
	
	set_timeslice(&sms_timeslice, sms_cpu);
    } else {
	deb_printf("out of memory.\n");
    }
}

/*
 * $Log: sms.c,v $
 * Revision 1.12  1999/01/10 19:55:38  nyef
 * added aliasing of bankswitch regs to 0xdffc-0xdfff.
 * added detection of sms/gg images.
 *
 * Revision 1.11  1999/01/10 02:20:19  nyef
 * disabled debug logging of writes to I/O ports 0xf0 and 0xf1.
 *
 * Revision 1.10  1999/01/09 23:10:06  nyef
 * added support for reading the current scanline.
 *
 * Revision 1.9  1999/01/09 22:57:01  nyef
 * added aliased I/O ports for controller reads.
 *
 * Revision 1.8  1999/01/09 21:52:13  nyef
 * fixed stupid mistake in sms_pagesync().
 *
 * Revision 1.7  1999/01/08 19:35:58  nyef
 * fixed timing in sms_timeslice().
 *
 * Revision 1.6  1999/01/06 02:41:12  nyef
 * added code to prevent non-extant banks on rom from being swapped in.
 *
 * Revision 1.5  1999/01/05 05:40:19  nyef
 * hacked in primitive controller support.
 *
 * Revision 1.4  1999/01/03 22:07:39  nyef
 * added vdp emulation stuff.
 * added some other stuff too.
 *
 * Revision 1.3  1999/01/03 02:20:19  nyef
 * switched from CPUT_EMUZ80 to CPUT_MARATZ80.
 *
 * Revision 1.2  1999/01/02 06:40:50  nyef
 * added dummy I/O interface.
 *
 * Revision 1.1  1999/01/01 20:43:47  nyef
 * Initial revision
 *
 */
