/*
 * sms.c
 *
 * Sega Master System emulation.
 */

/* $Id: sms.c,v 1.3 1999/01/03 02:20:19 nyef Exp $ */

#include <stdlib.h>

#include "ui.h"
#include "system.h"
#include "cal.h"
#include "tool.h"
#include "emuz80.h"

#define SMS_RAMSIZE 0xa000 /* 0x2000 ram + 0x8000 cart ram */

unsigned char *sms_memory;
cal_cpu sms_cpu;

unsigned char *sms_rombase;
unsigned char *sms_bank0;
unsigned char *sms_bank1;
unsigned char *sms_bank2;

unsigned char sms_pageregs[4];

void sms_pagesync(void)
{
    sms_bank0 = &sms_rombase[sms_pageregs[1] << 14];
    sms_bank1 = &sms_rombase[sms_pageregs[2] << 14];
    if (sms_rombase[0] & 0x08) {
	sms_bank2 = &sms_memory[0x2000 + ((sms_pageregs[0] & 0x04) << 12)];
    } else {
	sms_bank2 = &sms_rombase[sms_pageregs[3] << 14];
    }
}

unsigned char sms_read_rom0(cal_cpu cpu, unsigned short addr)
{
    if (addr < 0x400) {
	return sms_rombase[addr];
    } else {
	return sms_bank0[addr];
    }
}

unsigned char sms_read_rom1(cal_cpu cpu, unsigned short addr)
{
    return sms_bank1[addr & 0x3fff];
}

unsigned char sms_read_rom2(cal_cpu cpu, unsigned short addr)
{
    return sms_bank2[addr & 0x3fff];
}

unsigned char sms_read_ram(cal_cpu cpu, unsigned short addr)
{
    if (addr >= 0xfffc) {
	return sms_pageregs[addr & 3];
    } else {
	return sms_memory[addr & 0x1fff];
    }
}

void sms_write_ignore(cal_cpu cpu, unsigned short addr, unsigned char data)
{
}

void sms_write_rom2(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    if (sms_pageregs[0] & 0x08) {
	sms_bank2[addr & 0x3fff] = data;
    }
}

void sms_write_ram(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    if (addr >= 0xfffc) {
	sms_pageregs[addr & 3] = data;
	sms_pagesync();
    } else {
	sms_memory[addr & 0x1fff] = data;
    }
}

memread_t sms_readtable[4] = {
    sms_read_rom0, sms_read_rom1, sms_read_rom2, sms_read_ram,
};

memwrite_t sms_writetable[4] = {
    sms_write_ignore, sms_write_ignore, sms_write_rom2, sms_write_ram,
};

unsigned char sms_io_read(cal_cpu cpu, unsigned short address)
{
    deb_printf("sms: I/O read 0x%02x.\n", address & 0xff);
    return 0;
}

void sms_io_write(cal_cpu cpu, unsigned short address, unsigned char data)
{
    deb_printf("sms: I/O write 0x%02x = 0x%02x.\n", address & 0xff, data);
}

void sms_timeslice(cal_cpu cpu)
{
    int i;

    for (i = 42; i && (system_flags == F_NONE); i--) {
	cpu->runfor(cpu, 42); /* FIXME: this is totally arbitrary. */
	cpu->run(cpu);
	if (system_flags & F_UNIMPL) {
	    deb_printf("Unimplimented Instruction.\n");
	    deb_printf("PC: 0x%04hx, opcode: 0x%02x.\n", cpu->data.d_emuz80->reg_PC, sms_readtable[(cpu->data.d_emuz80->reg_PC >> 14) & 3](cpu, cpu->data.d_emuz80->reg_PC));
	}
    }
}

void sms_run(int startup_deb)
{
    sms_memory = calloc(1, SMS_RAMSIZE);
    
    if (sms_memory && romimage) {
	if (romsize & 0x200) {
	    sms_rombase = romimage + 0x200;
	} else {
	    sms_rombase = romimage;
	}
	sms_pageregs[0] = 0;
	sms_pageregs[1] = 0;
	sms_pageregs[2] = 1;
	sms_pageregs[3] = 2;
	sms_pagesync();
    
	sms_cpu = cal_create(CPUT_MARATZ80);
	sms_cpu->setmmu(sms_cpu, 14, 3, sms_readtable, sms_writetable);
	sms_cpu->setiou(sms_cpu, sms_io_read, sms_io_write);
	sms_cpu->reset(sms_cpu);

	set_timeslice(&sms_timeslice, sms_cpu);
    } else {
	deb_printf("out of memory.\n");
    }
}

/*
 * $Log: sms.c,v $
 * Revision 1.3  1999/01/03 02:20:19  nyef
 * switched from CPUT_EMUZ80 to CPUT_MARATZ80.
 *
 * Revision 1.2  1999/01/02 06:40:50  nyef
 * added dummy I/O interface.
 *
 * Revision 1.1  1999/01/01 20:43:47  nyef
 * Initial revision
 *
 */
