/*
 * sms9918.c
 *
 * sms custom VDP emulation.
 */

/* $Id: sms9918.c,v 1.9 1999/01/09 01:31:58 nyef Exp $ */

#include <stdlib.h>
#include "sms9918.h"
#include "ui.h"
#include "video.h"


/* manifest constants and flag defines */

#define SMS_VDP_RAMSIZE 0x4000

#define TF_ADDRWRITE 1


/* Palette definition */

int sms_palbase_red[64] = {
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
};

int sms_palbase_green[64] = {
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
};

int sms_palbase_blue[64] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55,
    0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa,
    0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
};


/* port 0 (data) emulation */

unsigned char sms9918_readport0(sms9918 vdp)
{
    unsigned char retval;

    if (vdp->address & 0x4000) {
	retval = vdp->palette[(vdp->address++) & 0x1f];
	vdp->address &= 0x401f;
    } else {
	retval = vdp->readahead;
	vdp->readahead = vdp->memory[vdp->address++];
	vdp->address &= 0x3fff;
    }
    vdp->flags &= ~TF_ADDRWRITE;
    return retval;
}

void sms9918_writeport0(sms9918 vdp, unsigned char data)
{
    if (vdp->address & 0x4000) {
	vdp->palette[(vdp->address++) & 0x1f] = data;
	vdp->address &= 0x401f;
    } else {
	vdp->readahead = data;
	vdp->memory[vdp->address++] = data;
	vdp->address &= 0x3fff;
    }
    vdp->flags &= ~TF_ADDRWRITE;
}


/* port 1 (status/address/palette/registers) emulation */

unsigned char sms9918_readport1(sms9918 vdp)
{
    unsigned char retval;

    retval = vdp->status;
    vdp->status &= 0x3f;
    vdp->flags &= ~TF_ADDRWRITE;
    return retval;
}

void sms9918_writeport1(sms9918 vdp, unsigned char data)
{
    if (vdp->flags & TF_ADDRWRITE) {
	if (data & 0x80) {
	    if (data & 0x40) {
		vdp->address = (vdp->addrsave | (data << 8)) & 0x401f;
	    } else {
		vdp->regs[data & 15] = vdp->addrsave;
	    }
	} else {
	    vdp->address = (vdp->addrsave | (data << 8)) & 0x3fff;
	    if (!(data & 0x40)) {
		vdp->readahead = vdp->memory[vdp->address++];
		vdp->address &= 0x3fff;
	    }
	}
	vdp->flags &= ~TF_ADDRWRITE;
    } else {
	vdp->addrsave = data;
	vdp->flags |= TF_ADDRWRITE;
    }
}


/* video rendering */

static unsigned char pixel_unpack[256][8];

void sms9918_init_pixelpacker(void)
{
    int i;

    for (i = 0; i < 0x100; i++) {
	pixel_unpack[i][0] = (i & 0x80)? 1: 0;
	pixel_unpack[i][1] = (i & 0x40)? 1: 0;
	pixel_unpack[i][2] = (i & 0x20)? 1: 0;
	pixel_unpack[i][3] = (i & 0x10)? 1: 0;
	pixel_unpack[i][4] = (i & 0x08)? 1: 0;
	pixel_unpack[i][5] = (i & 0x04)? 1: 0;
	pixel_unpack[i][6] = (i & 0x02)? 1: 0;
	pixel_unpack[i][7] = (i & 0x01)? 1: 0;
    }
}

void sms9918_render_background(sms9918 vdp, unsigned char *cur_vbp, int behind)
{
    unsigned short *cur_pos;
    unsigned short cur_tile;
    unsigned char *tiledata;
    unsigned char pixbuf[8];
    int cur_scanline;
    int cur_pattern_line;
    int scroll;
    int finescroll;
    int i;
    int j;

    cur_scanline = vdp->cur_scanline + vdp->regs[9];
    while (cur_scanline > 223) {
	cur_scanline -= 224;
    }
    cur_pattern_line = cur_scanline & 7;
    cur_pos = (unsigned short *)&vdp->memory[(vdp->regs[2] & 14) << 10];
    cur_pos += (cur_scanline >> 3) << 5;
    if ((vdp->regs[0] & 0x40) && (cur_scanline < 16)) {
	scroll = 0;
	finescroll = 0;
    } else {
	scroll = 32 - ((vdp->regs[8] + 7) >> 3);
	finescroll = (8 - (vdp->regs[8] & 7)) & 7;
    }
    
    for (i = 0; i < 32; i++) {
	cur_tile = cur_pos[(i + scroll) & 0x1f];
	if (cur_tile & 0x400) {
	    tiledata = &vdp->memory[((cur_tile & 0x1ff) << 5) + ((7 - cur_pattern_line) << 2)];
	} else {
	    tiledata = &vdp->memory[((cur_tile & 0x1ff) << 5) + (cur_pattern_line << 2)];
	}
	if (cur_tile & 0x200) {
	    for (j = 0; j < 8; j++) {
		pixbuf[7-j] = pixel_unpack[tiledata[0]][j];
		pixbuf[7-j] |= pixel_unpack[tiledata[1]][j] << 1;
		pixbuf[7-j] |= pixel_unpack[tiledata[2]][j] << 2;
		pixbuf[7-j] |= pixel_unpack[tiledata[3]][j] << 3;
	    }
	} else {
	    for (j = 0; j < 8; j++) {
		pixbuf[j] = pixel_unpack[tiledata[0]][j];
		pixbuf[j] |= pixel_unpack[tiledata[1]][j] << 1;
		pixbuf[j] |= pixel_unpack[tiledata[2]][j] << 2;
		pixbuf[j] |= pixel_unpack[tiledata[3]][j] << 3;
	    }
	}

	if ((!(cur_tile & 0x1000)) == (!behind)) {
	    cur_vbp += 8;
	} else {
	    if (i == 0) {
		if (cur_tile & 0x800) {
		    for (j = finescroll; j < 8; j++) {
		    *cur_vbp++ = vid_pre_xlat[vdp->palette[pixbuf[j] + 16]];
		    }
		} else {
		    for (j = finescroll; j < 8; j++) {
			*cur_vbp++ = vid_pre_xlat[vdp->palette[pixbuf[j]]];
		    }
		}
	    } else {
		if (cur_tile & 0x800) {
		    for (j = 0; j < 8; j++) {
			*cur_vbp++ = vid_pre_xlat[vdp->palette[pixbuf[j] + 16]];
		    }
		} else {
		    for (j = 0; j < 8; j++) {
			*cur_vbp++ = vid_pre_xlat[vdp->palette[pixbuf[j]]];
		    }
		}
	    }
	}
    }
}

void sms9918_render_sprites(sms9918 vdp, unsigned char *cur_vbp)
{
    unsigned char *sprite_table;
    unsigned char *tiledata;
    unsigned char pixbuf[8][8];
    unsigned short cur_tile;
    unsigned char sprite_x[8];
    int cur_pattern_line;
    int sprite_8x16;
    int i;
    int j;
    int k;

    k = 0;
    sprite_table = &vdp->memory[(vdp->regs[5] & 0x7e) << 7];
    sprite_8x16 = (vdp->regs[1] & 2)? 16: 8;
    
    for (i = 0; (i < 64) && (k < 8); i++) {
	if ((vdp->cur_scanline >= sprite_table[i]) &&
	    (vdp->cur_scanline < (sprite_table[i] + sprite_8x16))) {
	    cur_pattern_line = vdp->cur_scanline - sprite_table[i];

	    cur_tile = sprite_table[128 + (i << 1) + 1] | ((vdp->regs[6] & 4) << 6);
	    if (vdp->regs[1] & 2) {
		cur_tile &= 0x1fe;
	    }
	    tiledata = &vdp->memory[(cur_tile << 5) + (cur_pattern_line << 2)];
	    
	    for (j = 0; j < 8; j++) {
		pixbuf[k][j] = pixel_unpack[tiledata[0]][j];
		pixbuf[k][j] |= pixel_unpack[tiledata[1]][j] << 1;
		pixbuf[k][j] |= pixel_unpack[tiledata[2]][j] << 2;
		pixbuf[k][j] |= pixel_unpack[tiledata[3]][j] << 3;
	    }

	    sprite_x[k] = sprite_table[128 + (i << 1) + 0];
	    k++;
	} else if (sprite_table[i] == 208) {
	    break;
	}
    }
    for (i = 0; i < k; i++) {
	for (j = 0; j < 8; j++) {
	    if (pixbuf[i][j]) {
		cur_vbp[sprite_x[i] + j] = vid_pre_xlat[vdp->palette[pixbuf[i][j] + 16]];
	    }
	}
    }
}

void sms9918_render_line(sms9918 vdp)
{
    unsigned char *video_buffer;

    video_buffer = video_get_vbp(vdp->cur_scanline);
    
    if (vdp->regs[1] & 0x40) {
	sms9918_render_background(vdp, video_buffer, 1);
	sms9918_render_sprites(vdp, video_buffer);
	sms9918_render_background(vdp, video_buffer, 0);
    } else {
	/* FIXME: floodfill with black */
    }
}


/* periodic task */

int sms9918_periodic(sms9918 vdp)
{
    /* FIXME: does not account for overscan. */
    if (vdp->cur_scanline < 192) {
	sms9918_render_line(vdp);
    } else if (vdp->cur_scanline == 192) {
	video_events();
	video_display_buffer();
	vdp->status |= 0x80;
    }
    if (vdp->cur_scanline == 261) {
	vdp->cur_scanline = 0;
	vdp->linecounter = vdp->regs[10];
    } else {
	vdp->cur_scanline++;
	if (!--vdp->linecounter) {
	    vdp->status |= 0x40;
	    vdp->linecounter = vdp->regs[10];
	}
    }
    return (((vdp->status & 0x80) && (vdp->regs[1] & 0x20)) ||
	    ((vdp->status & 0x40) && (vdp->regs[0] & 0x10)));
}


/* initialization */

sms9918 sms9918_create(void)
{
    sms9918 retval;

    sms9918_init_pixelpacker();
    
    retval = calloc(1, sizeof(struct sms9918));
    if (retval) {
	retval->memory = calloc(1, SMS_VDP_RAMSIZE);
	if (retval->memory) {
	    retval->readport0 = sms9918_readport0;
	    retval->readport1 = sms9918_readport1;
	    retval->writeport0 = sms9918_writeport0;
	    retval->writeport1 = sms9918_writeport1;
	    retval->periodic = sms9918_periodic;
	    video_setpal(64, sms_palbase_red, sms_palbase_green, sms_palbase_blue);
	} else {
	    free(retval);
	    retval = NULL;
	}
    }
    
    if (!retval) {
	deb_printf("sms9918_create(): out of memory.\n");
    }
    
    return retval;
}

/*
 * $Log: sms9918.c,v $
 * Revision 1.9  1999/01/09 01:31:58  nyef
 * added support for 8x16 sprites.
 * limited number of sprites per line to 8 (supposedly the real system
 * does this too.).
 *
 * Revision 1.8  1999/01/09 00:48:35  nyef
 * fixed stupid mistake in tile priority code.
 *
 * Revision 1.7  1999/01/09 00:45:47  nyef
 * added preliminary support for vscroll.
 *
 * Revision 1.6  1999/01/09 00:39:12  nyef
 * added preliminary implimentation of tile priority flag.
 *
 * Revision 1.5  1999/01/08 19:36:21  nyef
 * improved timing in sms9918_periodic().
 *
 * Revision 1.4  1999/01/08 02:24:25  nyef
 * added line interrupts.
 *
 * Revision 1.3  1999/01/08 02:07:25  nyef
 * added horizontal scroll.
 *
 * Revision 1.2  1999/01/07 05:20:35  nyef
 * fixed sprite data table lookup.
 * added sprite tileset selection.
 * fixed sprite palette lookup.
 *
 * Revision 1.1  1999/01/07 03:17:07  nyef
 * Initial revision
 *
 */
