/*
 * torr_cpu.c
 *
 * CAL interface and non-generated code for torrigion core
 */

/* $Id: torr_cpu.c,v 1.2 1999/08/23 01:19:24 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include "cal.h"
#include "emu68k.h"
#include "tool.h"

#ifdef CPU_TORR68K

void cal_torr68k_reset(cal_cpu cpu);
void cal_torr68k_run(cal_cpu cpu);
void cal_torr68k_runfor(cal_cpu cpu, int cycles);
void cal_torr68k_irq(cal_cpu cpu);
void cal_torr68k_nmi(cal_cpu cpu);
void cal_torr68k_setzpage(cal_cpu cpu, void *page0);
int cal_torr68k_timeleft(cal_cpu cpu);
void cal_torr68k_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_torr68k_setmmu16(cal_cpu cpu, int shift, int mask, memread16_t *rtbl, memwrite16_t *wtbl);

void cal_torr68k_init(cal_cpu *cpu)
{
    (*cpu)->data.d_junk68k = calloc(1, sizeof(struct emu68k_context));
    if (!(*cpu)->data.d_junk68k) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_torr68k_reset;
    (*cpu)->run = cal_torr68k_run;
    (*cpu)->runfor = cal_torr68k_runfor;
    (*cpu)->irq = cal_torr68k_irq;
    (*cpu)->nmi = cal_torr68k_nmi;
    (*cpu)->setzpage = cal_torr68k_setzpage;
    (*cpu)->timeleft = cal_torr68k_timeleft;
    (*cpu)->setmmu8 = cal_torr68k_setmmu8;
    (*cpu)->setmmu16 = cal_torr68k_setmmu16;
    (*cpu)->data.d_junk68k->cpu = *cpu;
}

void cal_torr68k_reset(cal_cpu cpu)
{
    emu68k_reset(cpu->data.d_junk68k);
}

typedef void (*m68k_instr_handler)(struct emu68k_context *, unsigned short);

void cal_torr68k_run(cal_cpu cpu)
{
    unsigned short opword;
    struct emu68k_context *context;
    extern m68k_instr_handler torrigion_decode[0x10000];

    context = cpu->data.d_junk68k;
    opword = context->read16table[(context->pc >> context->memshift) & context->memmask](context->cpu, context->pc);
    context->pc += 2;
    if (torrigion_decode[opword]) {
	printf("torrigion: executing opword 0x%04hx.\n", opword);
	torrigion_decode[opword](cpu->data.d_junk68k, opword);
    } else {
	printf("torrigion: unimplemented opword 0x%04hx at 0x%06lx.\n", opword, context->pc);
	system_flags |= F_UNIMPL;
    }
}

void cal_torr68k_runfor(cal_cpu cpu, int cycles)
{
/*     cpu->data.d_emuz80->cycles_left += cycles; */
}

void cal_torr68k_irq(cal_cpu cpu)
{
/*     emuz80_IRQ(cpu->data.d_emuz80); */
}

void cal_torr68k_nmi(cal_cpu cpu)
{
/*     emuz80_NMI(cpu->data.d_emuz80); */
}

void cal_torr68k_setzpage(cal_cpu cpu, void *page0)
{
}

int cal_torr68k_timeleft(cal_cpu cpu)
{
/*     return cpu->data.d_emuz80->cycles_left; */
    return 0;
}

void cal_torr68k_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
    cpu->data.d_junk68k->memshift = shift;
    cpu->data.d_junk68k->memmask = mask;
    cpu->data.d_junk68k->read8table = rtbl;
    cpu->data.d_junk68k->write8table = wtbl;
}

void cal_torr68k_setmmu16(cal_cpu cpu, int shift, int mask, memread16_t *rtbl, memwrite16_t *wtbl)
{
    cpu->data.d_junk68k->read16table = rtbl;
    cpu->data.d_junk68k->write16table = wtbl;
}

#endif

/*
 * $Log: torr_cpu.c,v $
 * Revision 1.2  1999/08/23 01:19:24  nyef
 * fixed some include file problems
 *
 * Revision 1.1  1999/08/15 02:28:05  nyef
 * Initial revision
 *
 */
