/*
 * ui_allegro.c
 *
 * user interface for DJGPP/Allegro
 */

/* $Id: ui_allegro.c,v 1.4 1998/12/21 02:58:08 nyef Exp $ */

#include <stdio.h>
#include <stdarg.h>
#include "ui.h"
#include "system.h"
#include "video.h"
#include "tool.h"

void run_system(void);
shutdown_t shutdown;

int main(int argc, char *argv[])
{
    if (argc != 2) {
	printf("usage: %s <romfile>\n", argv[0]);
	exit(0x42);
    }
    
    video_init();

    read_romimage(argv[1]);
    activate_system(system_type, 0);

    run_system();

    return 0;
}


/* debug console handling */

void deb_printf(const char *fmt, ...)
{
#if 0
    va_list args;

    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);
#endif
}


/* emulation timeslicing */

void (*timeslice)(void *) = NULL;
void *timeslice_data;

void set_timeslice(void (*proc)(void *), void *data)
{
    timeslice = proc;
    timeslice_data = data;
}

void unset_timeslice(void)
{
    timeslice = NULL;
}

void run_system(void)
{
    while ((!(system_flags & F_QUIT)) && timeslice) {
	timeslice(timeslice_data);
    }
    shutdown();
}

/*
 * $Log: ui_allegro.c,v $
 * Revision 1.4  1998/12/21 02:58:08  nyef
 * added a shutdown callback.
 *
 * Revision 1.3  1998/12/18 04:17:14  nyef
 * changed usage message to use argv[0] for name of program.
 *
 * Revision 1.2  1998/08/04 01:57:59  nyef
 * fixed stipid mistake in main(). (forgot a 'return 0')
 *
 * Revision 1.1  1998/08/02 19:35:45  nyef
 * Initial revision
 *
 */
