/*
 * ui_allegro.c
 *
 * user interface for DJGPP/Allegro
 */

/* $Id: ui_allegro.c,v 1.7 1999/03/06 21:20:45 nyef Exp $ */

#include <stdio.h>
#include <stdarg.h>
#include <allegro.h>
#include "ui.h"
#include "system.h"
#include "video.h"
#include "tool.h"

void run_system(void);
shutdown_t shutdown;

int main(int argc, char *argv[])
{
    if (argc != 2) {
	printf("usage: %s <romfile>\n", argv[0]);
	exit(0x42);
    }
    
    video_init();

    read_romimage(argv[1]);
    activate_system(system_type, 0);

    run_system();

    return 0;
}


/* debug console handling */

void deb_printf(const char *fmt, ...)
{
#if 0
    va_list args;

    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);
#endif
}


/* input handling */

extern long joy1_data;

void video_events(void)
{
#if 0 /* this locks up the system. Oops. */
    if (key[KEY_F1]) {
	system_flags |= F_BREAK;
    } else {
	system_flags &= ~F_BREAK;
    }
#endif
    
    if (key[KEY_ESC]) {
	system_flags |= F_QUIT;
    } else {
	system_flags &= ~F_QUIT;
    }

    if (key[KEY_S]) {
	joy1_data |= 0x01;
    } else {
	joy1_data &= 0xfe;
    }
    
    if (key[KEY_A]) {
	joy1_data |= 0x02;
    } else {
	joy1_data &= 0xfd;
    }
    
    if (key[KEY_OPENBRACE]) {
	joy1_data |= 0x04;
    } else {
	joy1_data &= 0xfb;
    }
    
    if (key[KEY_CLOSEBRACE]) {
	joy1_data |= 0x08;
    } else {
	joy1_data &= 0xf7;
    }
    
    if (key[KEY_UP]) {
	joy1_data |= 0x10;
    } else {
	joy1_data &= 0xef;
    }
    
    if (key[KEY_DOWN]) {
	joy1_data |= 0x20;
    } else {
	joy1_data &= 0xdf;
    }
    
    if (key[KEY_LEFT]) {
	joy1_data |= 0x40;
    } else {
	joy1_data &= 0xbf;
    }
    
    if (key[KEY_RIGHT]) {
	joy1_data |= 0x80;
    } else {
	joy1_data &= 0x7f;
    }
}


/* emulation timeslicing */

void (*timeslice)(void *) = NULL;
void *timeslice_data;

void set_timeslice(void (*proc)(void *), void *data)
{
    timeslice = proc;
    timeslice_data = data;
}

void unset_timeslice(void)
{
    timeslice = NULL;
}

void run_system(void)
{
    while ((!(system_flags & F_QUIT)) && timeslice) {
	video_events();
	timeslice(timeslice_data);
    }
    if (shutdown) {
	shutdown();
    }
}

/*
 * $Log: ui_allegro.c,v $
 * Revision 1.7  1999/03/06 21:20:45  nyef
 * added inclusion of allegro.h (oops)
 *
 * Revision 1.6  1999/02/09 02:11:13  nyef
 * moved video_events() in from video_allegro.c
 * added call to video_events in main timeslice loop
 *
 * Revision 1.5  1999/01/01 20:50:51  nyef
 * fixed shutdown callback to only be called if it gets set up.
 *
 * Revision 1.4  1998/12/21 02:58:08  nyef
 * added a shutdown callback.
 *
 * Revision 1.3  1998/12/18 04:17:14  nyef
 * changed usage message to use argv[0] for name of program.
 *
 * Revision 1.2  1998/08/04 01:57:59  nyef
 * fixed stipid mistake in main(). (forgot a 'return 0')
 *
 * Revision 1.1  1998/08/02 19:35:45  nyef
 * Initial revision
 *
 */
