/*
 * ui_svga.c
 *
 * user interface for SVGALib
 */

/* $Id: ui_svga.c,v 1.1 1999/01/18 22:08:40 nyef Exp $ */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "ui.h"
#include "system.h"
#include "video.h"
#include "tool.h"

#include <vga.h>

#define TEMP_FILENAME "/tmp/darc.temp"

char *filename;

void run_system(void);
shutdown_t shutdown;

#define MAX_SVGA_MODE 2
struct {
    int modenumber;
    char name[12];
} modelist[MAX_SVGA_MODE] = {
    {G640x480x256,  "640x480x8"},
#ifdef G320x240x256V
    {G320x240x256V, "320x240x8"},  /* requires v1.4.0 of SVGALib */
#else
    {50,            "320x240x8"},
#endif
};

extern int svga_mode;


int fexists(char *fn)
{
    FILE *F = fopen(fn,"rb");
    
    if (F==NULL) return 0;
    fclose(F);
    return 1;
}

void printusage(char *argv[])
{
    printf("usage: %s <romfile> <-m #> <-h>\n", argv[0]);
}

void listmodes()
{
    int x;

    printf("\nVideo modes:\n");
    for(x=0; x<MAX_SVGA_MODE; x++) {
	printf("    %d - %s %s\n", x, modelist[x].name,
	       vga_hasmode(modelist[x].modenumber)?"":"(not available)");
    }
}

void checkargs(int argc, char *argv[])
{
    int x;
    
    if (argc == 1) {
	printusage(argv);
	exit(0x42);
    }
    
    for (x=1; x<argc; x++) {
	/* Check if it's a filename */
	if (argv[x][0] != '-') {
	    if (filename == NULL) { /* Do we already have a filename? */
		if (!fexists(argv[x])) {
		    fprintf(stderr, "%s not found\n", argv[x]);
		    exit(-1);
		} else {
		    filename = strdup(argv[x]); /* Yay! */
		}
	    }
	} else { /* It's a switch! ----------------------- */
	    switch(argv[x][1]) {
	    case 'm':
		x++;
		if (x > argc) {
		    printf("Bad argument\n");
		    exit(-1);
		}
		svga_mode = atoi(argv[x]);
		if ((svga_mode < 0) || (svga_mode >= MAX_SVGA_MODE)) {
		    printf("Video mode undefined.\n");
		    exit(-1);
		}
		svga_mode = modelist[svga_mode].modenumber;
		break;
	    case 'h': /* Help */
		printusage(argv);
		listmodes();
		exit(0);
		break; /* etc... */
	    }
	}
    }
    if (filename == NULL) {
	printusage(argv);
	exit(0x42);
    }
}

int main(int argc, char *argv[])
{
    checkargs(argc, argv);
    
    if (strstr(filename, ".gz\0")) { /* If file looks gzip compressed... */
	char temp[256];    /* Lets unpack it into the /tmp dir... */
	
	sprintf(temp, "gunzip -c %s >"TEMP_FILENAME, filename);
	if (system(temp)) {
	    printf("Error gunzipping compressed rom.\n");
	    exit(-1);
        }
	read_romimage(TEMP_FILENAME);
	unlink(TEMP_FILENAME);
    } else {             /* ...otherwise, business as usual. */
	read_romimage(filename);
    }
    free(filename);

    video_init();
    activate_system(system_type, 0);
    run_system();
    
    return 0;
}

/* debug console handling */
void deb_printf(const char *fmt, ...)
{
#if 0
    va_list args;

    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);
#endif
}

/* emulation timeslicing */

void (*timeslice)(void *) = NULL;
void *timeslice_data;

void set_timeslice(void (*proc)(void *), void *data)
{
    timeslice = proc;
    timeslice_data = data;
}

void unset_timeslice(void)
{
    timeslice = NULL;
}

void run_system(void)
{
    while ((!(system_flags & F_QUIT)) && timeslice) {
	timeslice(timeslice_data);
    }
    
    if (shutdown) {
	shutdown();
    }
}

/*
 * $Log: ui_svga.c,v $
 * Revision 1.1  1999/01/18 22:08:40  nyef
 * Initial revision
 *
 */
