/*
 * ui_x.c
 *
 * user interface for the X Window System.
 */

/* $Id: ui_x.c,v 1.9 1998/12/21 02:58:03 nyef Exp $ */

#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>

#include <stdio.h>
#include <stdarg.h>
#include "ui.h"
#include "system.h"
#include "video.h"
#include "tool.h"

/* Overhead for Xt and main window */
XtAppContext context;
Widget toplevel;
Widget manager;

/* Menus */
Widget menubar;
Widget menubutton_file;
Widget menu_file;
Widget menu_file_exit;
Widget menubutton_debug;
Widget menu_debug;
Widget menu_debug_console;
Widget menu_debug_regs;
Widget menu_debug_disas;

/* the drawing area */
Widget drawbox;

/* the code window */
Widget code_shell;
Widget code_manager;
Widget linelabels[16];

/* the shutdown callback */
shutdown_t shutdown;

/* miscellany */
void do_file_exit(Widget, XtPointer, XtPointer);
void do_debug_console(Widget, XtPointer, XtPointer);
void do_debug_regs(Widget, XtPointer, XtPointer);
void do_debug_disas(Widget, XtPointer, XtPointer);
Boolean emulation_timeslice(XtPointer);

void init_menus(void)
{
    /* the menubar */
    menubar = XtVaCreateManagedWidget("menubar", boxWidgetClass, manager, XtNorientation, XtorientHorizontal, XtNborderWidth, 0, NULL);

    /* the file menu */
    menubutton_file = XtCreateManagedWidget("File", menuButtonWidgetClass, menubar, NULL, 0);
    menu_file = XtCreatePopupShell("fileMenu", simpleMenuWidgetClass, menubutton_file, NULL, 0);
    XtVaSetValues(menubutton_file, XtNmenuName, "fileMenu", NULL);
    menu_file_exit = XtCreateManagedWidget("Exit", smeBSBObjectClass, menu_file, NULL, 0);
    XtAddCallback(menu_file_exit, XtNcallback, do_file_exit, NULL);

    /* the debug menu */
    menubutton_debug = XtCreateManagedWidget("Debug", menuButtonWidgetClass, menubar, NULL, 0);
    menu_debug = XtCreatePopupShell("debugMenu", simpleMenuWidgetClass, menubutton_debug, NULL, 0);
    XtVaSetValues(menubutton_debug, XtNmenuName, "debugMenu", NULL);
    menu_debug_console = XtCreateManagedWidget("Show console", smeBSBObjectClass, menu_debug, NULL, 0);
    XtAddCallback(menu_debug_console, XtNcallback, do_debug_console, NULL);
    menu_debug_regs = XtCreateManagedWidget("Dump regs", smeBSBObjectClass, menu_debug, NULL, 0);
    XtAddCallback(menu_debug_regs, XtNcallback, do_debug_regs, NULL);
    menu_debug_disas = XtCreateManagedWidget("Disassemble", smeBSBObjectClass, menu_debug, NULL, 0);
    XtAddCallback(menu_debug_disas, XtNcallback, do_debug_disas, NULL);
}

void init_codeshell(void)
{
    int i;
    char buf[12];
    
    code_shell = XtVaCreatePopupShell("codeShell", topLevelShellWidgetClass, manager, NULL);
    code_manager = XtVaCreateManagedWidget("codeManager", boxWidgetClass, code_shell, XtNhSpace, 0, XtNvSpace, 0, NULL);

    for (i = 0; i < 16; i++) {
	sprintf(buf, "linelabel%d\n", i);
	linelabels[i] = XtVaCreateManagedWidget(buf, labelWidgetClass, code_manager, XtNlabel, "insert code here", NULL);
    }
}

int main(int argc, char *argv[])
{
    toplevel = XtAppInitialize(&context, "DarcNES", NULL, 0, &argc, argv, NULL, NULL, 0);
    manager = XtVaCreateManagedWidget("manager", boxWidgetClass, toplevel, XtNhSpace, 0, XtNvSpace, 0, NULL);

    init_menus();
    
    drawbox = XtVaCreateManagedWidget("drawbox", boxWidgetClass, manager, XtNwidth, 256, XtNheight, 240, NULL);

    init_codeshell();

    XtRealizeWidget(toplevel);
    
    video_init();
    
    read_romimage("cart.nes");
    activate_system(system_type, 0);

    XtAppAddWorkProc(context, emulation_timeslice, NULL);

    XtAppMainLoop(context);
    
    return 0;
}


/* debug console handling */

void deb_printf(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);
}


/* emulation timeslicing */

void (*timeslice)(void *) = NULL;
void *timeslice_data;

void set_timeslice(void (*proc)(void *), void *data)
{
    timeslice = proc;
    timeslice_data = data;
}

void unset_timeslice(void)
{
    timeslice = NULL;
}

Boolean emulation_timeslice(XtPointer client_data)
{
    int i;

    for (i = 243; i && (system_flags == F_NONE); i--) {
	if (timeslice) {
	    timeslice(timeslice_data);
	}
    }
    if (system_flags == F_NONE) {
	return False;
    } else {
	return True;
    }
}


/* menu callbacks */

void do_file_exit(Widget w, XtPointer client_data, XtPointer call_data)
{
    system_flags |= F_QUIT;
    shutdown();
}

void do_debug_console(Widget w, XtPointer client_data, XtPointer call_data)
{
    printf("do_debug_console()\n");
    XtPopup(code_shell, XtGrabNone);
}

#include "debug.h"
#include "emu6502.h"

void do_debug_regs(Widget w, XtPointer client_data, XtPointer call_data)
{
    dump_regs(timeslice_data);
}

void do_debug_disas(Widget w, XtPointer client_data, XtPointer call_data)
{
    disassemble(((struct emu6502_context *)timeslice_data)->reg_PC, ((((struct emu6502_context *)timeslice_data)->reg_PC+16) & 0xffff));
}


/* video interface */

void ui_set_drawbox_size(int width, int height)
{
    XtVaSetValues(drawbox, XtNwidth, width, XtNheight, height, NULL);
}

/*
 * $Log: ui_x.c,v $
 * Revision 1.9  1998/12/21 02:58:03  nyef
 * added a shutdown callback.
 *
 * Revision 1.8  1998/12/18 04:22:25  nyef
 * changed application class to "DarcNES".
 *
 * Revision 1.7  1998/10/28 01:50:00  nyef
 * changed emulation_timeslice() to return aproximately once per screen
 * refresh. this sped up the emulation appreciably but does not appear
 * to have degraded the key response.
 *
 * Revision 1.6  1998/10/02 01:22:07  nyef
 * changed timeslice function to remove itself when emulation halts.
 *
 * Revision 1.5  1998/08/24 01:25:41  nyef
 * started working on new interface.
 *
 * Revision 1.4  1998/08/02 04:01:56  nyef
 * added function deb_printf() for debug console support.
 * added ui.h to the list of includes.
 *
 * Revision 1.3  1998/08/02 03:56:28  nyef
 * added comments grouping functions into related sections.
 *
 * Revision 1.2  1998/08/02 02:31:47  nyef
 * added a separate menubar widget.
 * added a "Debug" menu with one item. ("Show console", which calls a stub.)
 *
 * Revision 1.1  1998/08/02 02:04:26  nyef
 * Initial revision
 *
 */
