/*
 * video_svga.c
 *
 * display screen management
 */

/* $Id: video_svga.c,v 1.2 1999/01/18 22:08:18 nyef Exp $ */

#include <vga.h>
#include <vgagl.h>
#include <vgakeyboard.h>
/* #include <vgajoystick.h> */

#include <stdio.h>
#include <stdlib.h>

#include "video.h"
#include "tool.h"

#define BUFFER_X 256
#define BUFFER_Y 240
#define BUFFER_SIZE (BUFFER_X * BUFFER_Y)


int video_active = 0;
int deblevel = 0;

unsigned char *vid_pre_xlat;
video_setpal_t video_setpal;
vbpfunc        video_get_vbp;

unsigned char *nes_image;

int middle_x=0,
    middle_y=0;

int svga_mode = G640x480x256;
int use_joystick = 1;
int joy_up, joy_down, joy_left, joy_right,
    joy_a, joy_b, joy_start, joy_select;

void init_xlate8(int colors, int *red, int *green, int *blue);
unsigned char *vbpfunc8(int line);

__inline__
void input_update()
{
    keyboard_update();
    
    if (keyboard_keypressed(SCANCODE_ESCAPE)) {
	system_flags |= F_QUIT;
    } else {
	system_flags &= ~F_QUIT;
    }
    
    joy_a = (keyboard_keypressed(SCANCODE_S) || keyboard_keypressed(SCANCODE_LEFTALT));
    joy_b = (keyboard_keypressed(SCANCODE_A) || keyboard_keypressed(SCANCODE_LEFTCONTROL));
    joy_select = (keyboard_keypressed(SCANCODE_BRACKET_LEFT) || keyboard_keypressed(SCANCODE_TAB));
    joy_start  = (keyboard_keypressed(SCANCODE_BRACKET_RIGHT) || keyboard_keypressed(SCANCODE_ENTER));
    joy_up     = keyboard_keypressed(SCANCODE_CURSORUP);
    joy_down   = keyboard_keypressed(SCANCODE_CURSORDOWN);
    joy_left   = keyboard_keypressed(SCANCODE_CURSORLEFT);
    joy_right  = keyboard_keypressed(SCANCODE_CURSORRIGHT);
    
#if 0
    if (use_joystick) {
	joystick_update();
	
	joy_up    = (joystick_x(0) > 0) ? 1 : joy_up;
	joy_down  = (joystick_x(0) < 0) ? 1 : joy_down;
	joy_left  = (joystick_y(0) < 0) ? 1 : joy_left;
	joy_right = (joystick_y(0) > 0) ? 1 : joy_right;
    }
#endif
}

void video_init()
{
    vga_disabledriverreport();
    vga_init();
    video_setpal = init_xlate8;
    
    if (!vga_hasmode(svga_mode)) {
	printf("This video mode requires a more recent version of\n"
	       "SVGALib, such as v1.4.0, which is available from\n"
               "http://www.cs.bgu.ac.il/~zivav/svgalib/\n");
        exit(-1);
    }
    vga_setmode(svga_mode);
    gl_setcontextvga(svga_mode);
    
    middle_x = (vga_getxdim() / 2) - (BUFFER_X/2);
    middle_y = (vga_getydim() / 2) - (BUFFER_Y/2);
    
    keyboard_init();
    keyboard_translatekeys(TRANSLATE_CURSORKEYS|TRANSLATE_KEYPADENTER|
                           TRANSLATE_DIAGONAL);

#if 0
    if (use_joystick) {
	if (joystick_init(0,NULL) < 0) {
	    fprintf(stderr, "nojoy\n");
	}
    }
#endif
    
    nes_image = (unsigned char *)malloc(BUFFER_SIZE);
    if (nes_image == NULL) {
	video_shutdown();
	printf("Error allocating %d bytes for a video buffer...\n", BUFFER_SIZE);
	exit(-1);
    }
    atexit(video_shutdown);
}

void video_shutdown(void)
{
    vga_setmode(TEXT);
    keyboard_close();
}

void video_enter_deb(void)
{
    if (video_active && (!deblevel)) {
	vga_flip();
    }
    deblevel++;
}

void video_leave_deb(void)
{
    deblevel--;
    if (video_active && (!deblevel)) {
	vga_flip();
    }
}

__inline__
void video_display_buffer()
{
    gl_putbox(middle_x,middle_y, BUFFER_X, BUFFER_Y, nes_image);
}

extern long joy1_data;

void video_events(void)
{
    input_update();

    if (joy_a) {
	joy1_data |= 0x01;     /* A BUTTON */
    } else {
	joy1_data &= 0xfe;
    }
    
    if (joy_b) {
	joy1_data |= 0x02;     /* B BUTTON */
    } else {
	joy1_data &= 0xfd;
    }
    
    if (joy_select) {
	joy1_data |= 0x04;
    } else {
	joy1_data &= 0xfb;
    }

    if (joy_start) {
	joy1_data |= 0x08; /* START */
    } else {
	joy1_data &= 0xf7;
    }
    
    if (joy_up) {
	joy1_data |= 0x10; /* UP */
    } else {
	joy1_data &= 0xef;
    }
    
    if (joy_down) {
	joy1_data |= 0x20; /* DOWN */
    } else {
	joy1_data &= 0xdf;
    }
    
    if (joy_left) {
	joy1_data |= 0x40; /* LEFT */
    } else {
	joy1_data &= 0xbf;
    }
    
    if (joy_right) {
	joy1_data |= 0x80; /* RIGHT */
    } else {
	joy1_data &= 0x7f;
    }
}

void init_xlate8(int colors, int *red, int *green, int *blue)
{
    int i;
    
    vid_pre_xlat = (unsigned char*) malloc(colors);
    video_get_vbp = vbpfunc8;
    
    for (i = 0; i < 64; i++) {
	gl_setpalettecolor(i+64, red[i] >> 2, green[i] >> 2, blue[i] >> 2);
	vid_pre_xlat[i] = i + 64;
    }
}


unsigned char *vbpfunc8(int line)
{
    return nes_image+(line*256);
}

/*
 * $Log: video_svga.c,v $
 * Revision 1.2  1999/01/18 22:08:18  nyef
 * added a cleaned up version of the changes by Toby Deshane.
 *
 * Revision 1.1  1998/07/11 22:19:21  nyef
 * Initial revision
 *
 */
