/*
 * video_x.c
 *
 * display screen management
 */

/* $Id: video_x.c,v 1.14 1999/02/13 22:23:51 nyef Exp $ */

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <stdio.h>
#include <stdlib.h>
#include "video.h"
#include "tool.h"

int video_active = 0;
int deblevel = 0;

Display *display;
Window nes_window;
GC nes_gc;
XImage *nes_image;
Colormap nes_cmap;
Screen *nes_screen;

int buffer_x;
int buffer_y;
char *image_buf;

typedef void ((*xlatfunc)(void *, int));

vbpfunc video_get_vbp;
unsigned char *vbpfunc8(int line);
unsigned char *vbpfunc16(int line);
xlatfunc vidxlate;
void vidxlate16 (char *dest, int size);
video_setpal_t video_setpal;
void init_xlate8(int colors, int *red, int *green, int *blue);
void init_xlate16(int colors, int *red, int *green, int *blue);
unsigned char *vid_pre_xlat;

/* FIXME: ui_x.c interface -- move to separate header. */
extern Widget drawbox;
extern Widget manager;
void ui_set_drawbox_size(int width, int height);

void do_key_event(Widget, XtPointer, XEvent *, Boolean *);

void video_init()
{
    display = XtDisplay(drawbox);
    nes_screen = XtScreen(drawbox);
    nes_window = XtWindow(drawbox);

    ui_set_drawbox_size(256, 240);

    nes_gc = DefaultGCOfScreen(nes_screen);
    XtVaGetValues(drawbox, XtNcolormap, &nes_cmap, NULL);

    XtAddEventHandler(manager, KeyPressMask | KeyReleaseMask, False, do_key_event, NULL);
    
    XFlush(display);
    
    if (DefaultDepthOfScreen(nes_screen) == 8) {
	video_setpal = init_xlate8;
    } else if (DefaultDepthOfScreen(nes_screen) == 16) {
	video_setpal = init_xlate16;
    } else {
	fprintf(stderr, "display depth %d unsupported.\n", DefaultDepthOfScreen(nes_screen));
	exit(1);
    }
}

void video_shutdown(void)
{
}

void video_setsize(int x, int y)
{
    buffer_x = x;
    buffer_y = y;

    ui_set_drawbox_size(x, y);
    
    if (DefaultDepthOfScreen(nes_screen) == 8) {
	image_buf = malloc(x * y);
	nes_image = XCreateImage(display, DefaultVisualOfScreen(nes_screen),
				 8, ZPixmap, 0, image_buf, x, y, 8, 0);
    } else if (DefaultDepthOfScreen(nes_screen) == 16) {
	image_buf = malloc(x * y * 2);
	nes_image = XCreateImage(display, DefaultVisualOfScreen(nes_screen),
				 16, ZPixmap, 0, image_buf, x, y, 8, 0);
    }
}

void video_enter_deb(void)
{
    if (video_active && (!deblevel)) {

    }
    deblevel++;
}

void video_leave_deb(void)
{
    deblevel--;
    if (video_active && (!deblevel)) {

    }
}

void video_display_buffer(void)
{
    if (vidxlate) {
	vidxlate(image_buf, buffer_x * buffer_y);
    }
    XPutImage(display, nes_window, nes_gc, nes_image, 0, 0, 0, 0, buffer_x, buffer_y);
    XFlush(display);
}

extern long joy1_data;

void video_events(void)
{
}

void do_key_event(Widget w, XtPointer client_data, XEvent *event, Boolean *continue_dispatch)
{
    KeySym keysym;
    Modifiers dummy;
    static int key_up_1;
    static int key_down_1;
    static int key_left_1;
    static int key_right_1;
    static int key_up_2;
    static int key_down_2;
    static int key_left_2;
    static int key_right_2;

    XtTranslateKeycode(display, event->xkey.keycode, event->xkey.state, &dummy, &keysym);
    
    if (event->type == KeyPress) {
	switch (keysym) {
	case XK_F1:
	    /* enter debugger */
	    system_flags |= F_BREAK;
	    break;
	    
	case XK_Escape:
	    /* quit */
	    system_flags |= F_QUIT;
	    break;
	case XK_s: case XK_S:
	    joy1_data |= 0x01;
	    break;
	case XK_a: case XK_A:
	    joy1_data |= 0x02;
	    break;
	case XK_bracketleft: case XK_braceleft:
	    joy1_data |= 0x04;
	    break;
	case XK_bracketright: case XK_braceright:
	    joy1_data |= 0x08;
	    break;
	case XK_Up:
	    key_up_1 = 1;
	    break;
	case XK_Down:
	    key_down_1 = 1;
	    break;
	case XK_Left:
	    key_left_1 = 1;
	    break;
	case XK_Right:
	    key_right_1 = 1;
	    break;
	case XK_KP_Up:
	    key_up_2 = 1;
	    break;
	case XK_KP_Down:
	    key_down_2 = 1;
	    break;
	case XK_KP_Left:
	    key_left_2 = 1;
	    break;
	case XK_KP_Right:
	    key_right_2 = 1;
	    break;
	}
    } else {
	switch (keysym) {
	case XK_s: case XK_S:
	    joy1_data &= ~0x01;
	    break;
	case XK_a: case XK_A:
	    joy1_data &= ~0x02;
	    break;
	case XK_bracketleft: case XK_braceleft:
	    joy1_data &= ~0x04;
	    break;
	case XK_bracketright: case XK_braceright:
	    joy1_data &= ~0x08;
	    break;
	case XK_Up:
	    key_up_1 = 0;
	    break;
	case XK_Down:
	    key_down_1 = 0;
	    break;
	case XK_Left:
	    key_left_1 = 0;
	    break;
	case XK_Right:
	    key_right_1 = 0;
	    break;
	case XK_KP_Up:
	    key_up_2 = 0;
	    break;
	case XK_KP_Down:
	    key_down_2 = 0;
	    break;
	case XK_KP_Left:
	    key_left_2 = 0;
	    break;
	case XK_KP_Right:
	    key_right_2 = 0;
	    break;
	}
    }

    joy1_data &= 0x0f;
    
    if (key_up_1 || key_up_2) {
	joy1_data |= 0x10;
    }
    if (key_down_1 || key_down_2) {
	joy1_data |= 0x20;
    }
    if (key_left_1 || key_left_2) {
	joy1_data |= 0x40;
    }
    if (key_right_1 || key_right_2) {
	joy1_data |= 0x80;
    }
}

short *xlatepal;

void vidxlate16(char *dest, int size)
{
    int i;

    for (i = size; i > 0; i--) {
	*(((short *)dest)+(i-1)) = xlatepal[*(dest+(i-1))];
    }
}

unsigned char *vbpfunc8(int line)
{
    return image_buf + (line * buffer_x);
}

unsigned char *vbpfunc16(int line)
{
    return image_buf + (line * buffer_x);
}

/* #include "nespal.c" */

void init_xlate8(int colors, int *red, int *green, int *blue)
{
    int i;
    XColor color;
    
    vid_pre_xlat = malloc(colors);
    vidxlate = NULL;
    video_get_vbp = vbpfunc8;
    
    for (i = 0; i < colors; i++) {
	color.flags = DoRed | DoGreen | DoBlue;
	color.red = red[i];
	color.green = green[i];
	color.blue = blue[i];
	color.red += color.red << 8;
	color.green += color.green << 8;
	color.blue += color.blue << 8;
	if (XAllocColor(display, nes_cmap, &color)) {
	    vid_pre_xlat[i] = color.pixel;
	}
    }
}

void init_xlate16(int colors, int *red, int *green, int *blue)
{
    int i;
    XColor color;

    xlatepal = malloc(colors << 1);
    vid_pre_xlat = malloc(colors);
    vidxlate = (xlatfunc)&vidxlate16;
    video_get_vbp = vbpfunc16;
    
    for (i = 0; i < colors; i++) {
	vid_pre_xlat[i] = i;
	color.flags = DoRed | DoGreen | DoBlue;
	color.red = red[i];
	color.green = green[i];
	color.blue = blue[i];
	color.red += color.red << 8;
	color.green += color.green << 8;
	color.blue += color.blue << 8;
	if (XAllocColor(display, nes_cmap, &color)) {
	    xlatepal[i] = color.pixel;
	}
    }
}

/*
 * $Log: video_x.c,v $
 * Revision 1.14  1999/02/13 22:23:51  nyef
 * added support for the keypad arrow keys
 *
 * Revision 1.13  1999/02/07 17:55:51  nyef
 * removed double-buffering code
 * added video_setsize() interface
 *
 * Revision 1.12  1999/01/05 04:28:41  nyef
 * hacked up an interface to allow setting the palette.
 *
 * Revision 1.11  1998/08/29 22:09:37  nyef
 * switched to a double-buffered video system. added routines to support
 * the PPU writing directly into the video buffers. eliminated the 8-bit
 * vidxlate routine. changed vidxlate16 to operate using only one buffer.
 * changed the video buffer size to a symbolic constant.
 *
 * Revision 1.10  1998/08/29 14:52:29  nyef
 * added an identity vid_pre_xlat table for use in 16-bitplane modes.
 *
 * Revision 1.9  1998/08/28 03:02:16  nyef
 * added a "vid_pre_xlat" table so that some of the color translation
 * could be done by the PPU. also moved the 8-bit color translation to
 * the PPU and set the 8-bit xlatfinc to be memcpy. (maybe I should
 * double buffer the display?)
 *
 * Revision 1.8  1998/08/15 15:45:34  nyef
 * reworked vidxlate16 to use a normal loop and to copy longs.
 * this should be slightly faster on a pentium.
 *
 * Revision 1.7  1998/08/15 01:08:15  nyef
 * changed unsupported display depth message to indicate current depth.
 *
 * Revision 1.6  1998/08/02 02:03:08  nyef
 * preliminary GUIfication completed.
 *
 * Revision 1.5  1998/08/01 23:12:45  nyef
 * removed the checks for the symbol "GRAPHICS".
 * removed two commented-out printf statements in the input handler.
 *
 * Revision 1.4  1998/08/01 00:57:13  nyef
 * moved old palette values into nespal.c
 *
 * Revision 1.3  1998/07/26 13:22:55  nyef
 * changed vidxlate16 tu use pointer notation and duff's device.
 *
 * Revision 1.2  1998/07/18 18:54:27  nyef
 * changed vidxlate16 to copy shorts instead of chars for a massive speed increase
 *   in 16 bitplane mode.
 *
 * Revision 1.1  1998/07/11 22:19:25  nyef
 * Initial revision
 *
 */
