/*
 * video_x.c
 *
 * display screen management
 */

/* $Id: video_x.c,v 1.11 1998/08/29 22:09:37 nyef Exp $ */

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <stdio.h>
#include <stdlib.h>
#include "video.h"
#include "tool.h"

int video_active = 0;
int deblevel = 0;

Display *display;
Window nes_window;
GC nes_gc;
XImage *nes_image1;
XImage *nes_image2;
XImage *cur_image;
Colormap nes_cmap;

#define VIDEO_BUFFER_SIZE (256*240)
char *image_buf1;
char *image_buf2;
char *cur_buf;

typedef void ((*xlatfunc)(void *, int));

vbpfunc video_get_vbp;
unsigned char *vbpfunc8(int line);
unsigned char *vbpfunc16(int line);
xlatfunc vidxlate;
void vidxlate16 (char *dest, int size);
void init_xlate8(void);
void init_xlate16(void);
unsigned char *vid_pre_xlat;

/* FIXME: ui_x.c interface -- move to separate header. */
extern Widget drawbox;
extern Widget manager;
void ui_set_drawbox_size(int width, int height);

void do_key_event(Widget, XtPointer, XEvent *, Boolean *);

void video_init()
{
    Screen *screen;
    
    display = XtDisplay(drawbox);
    screen = XtScreen(drawbox);
    nes_window = XtWindow(drawbox);

    ui_set_drawbox_size(256, 240);

    nes_gc = DefaultGCOfScreen(screen);
    XtVaGetValues(drawbox, XtNcolormap, &nes_cmap, NULL);

    XtAddEventHandler(manager, KeyPressMask | KeyReleaseMask, False, do_key_event, NULL);
    
    XFlush(display);
    
    if (DefaultDepthOfScreen(screen) == 8) {
	image_buf1 = malloc(VIDEO_BUFFER_SIZE);
	nes_image1 = XCreateImage(display, DefaultVisualOfScreen(screen), 8,
				 ZPixmap, 0, image_buf1, 256, 240, 8, 0);
	image_buf2 = malloc(VIDEO_BUFFER_SIZE);
	nes_image2 = XCreateImage(display, DefaultVisualOfScreen(screen), 8,
				 ZPixmap, 0, image_buf2, 256, 240, 8, 0);
	init_xlate8();
    } else if (DefaultDepthOfScreen(screen) == 16) {
	image_buf1 = malloc(VIDEO_BUFFER_SIZE * 2);
	nes_image1 = XCreateImage(display, DefaultVisualOfScreen(screen), 16,
				 ZPixmap, 0, image_buf1, 256, 240, 8, 0);
	image_buf2 = malloc(VIDEO_BUFFER_SIZE * 2);
	nes_image2 = XCreateImage(display, DefaultVisualOfScreen(screen), 16,
				 ZPixmap, 0, image_buf2, 256, 240, 8, 0);
	init_xlate16();
    } else {
	fprintf(stderr, "display depth %d unsupported.\n", DefaultDepthOfScreen(screen));
	exit(1);
    }

    cur_image = nes_image2;
    cur_buf = image_buf1;
}

void video_shutdown(void)
{
}

void video_enter_deb(void)
{
    if (video_active && (!deblevel)) {

    }
    deblevel++;
}

void video_leave_deb(void)
{
    deblevel--;
    if (video_active && (!deblevel)) {

    }
}

void video_display_buffer(void)
{
    if (vidxlate) {
	vidxlate(cur_buf, VIDEO_BUFFER_SIZE);
    }
    XPutImage(display, nes_window, nes_gc, cur_image, 0, 0, 0, 0, 256, 240);
    XFlush(display);
    cur_image = (cur_image == nes_image1)? nes_image2: nes_image1;
    cur_buf = (cur_buf == image_buf1)? image_buf2: image_buf1;
}

extern long joy1_data;

void video_events(void)
{
}

void do_key_event(Widget w, XtPointer client_data, XEvent *event, Boolean *continue_dispatch)
{
    KeySym keysym;
    Modifiers dummy;

    XtTranslateKeycode(display, event->xkey.keycode, event->xkey.state, &dummy, &keysym);
    
    if (event->type == KeyPress) {
	switch (keysym) {
	case XK_F1:
	    /* enter debugger */
	    system_flags |= F_BREAK;
	    break;
	    
	case XK_Escape:
	    /* quit */
	    system_flags |= F_QUIT;
	    break;
	case XK_s: case XK_S:
	    joy1_data |= 0x01;
	    break;
	case XK_a: case XK_A:
	    joy1_data |= 0x02;
	    break;
	case XK_bracketleft: case XK_braceleft:
	    joy1_data |= 0x04;
	    break;
	case XK_bracketright: case XK_braceright:
	    joy1_data |= 0x08;
	    break;
	case XK_Up:
	    joy1_data |= 0x10;
	    break;
	case XK_Down:
	    joy1_data |= 0x20;
	    break;
	case XK_Left:
	    joy1_data |= 0x40;
	    break;
	case XK_Right:
	    joy1_data |= 0x80;
	    break;
	}
    } else {
	switch (keysym) {
	case XK_s: case XK_S:
	    joy1_data &= ~0x01;
	    break;
	case XK_a: case XK_A:
	    joy1_data &= ~0x02;
	    break;
	case XK_bracketleft: case XK_braceleft:
	    joy1_data &= ~0x04;
	    break;
	case XK_bracketright: case XK_braceright:
	    joy1_data &= ~0x08;
	    break;
	case XK_Up:
	    joy1_data &= ~0x10;
	    break;
	case XK_Down:
	    joy1_data &= ~0x20;
	    break;
	case XK_Left:
	    joy1_data &= ~0x40;
	    break;
	case XK_Right:
	    joy1_data &= ~0x80;
	    break;
	}
    }
}

short *xlatepal;

void vidxlate16(char *dest, int size)
{
    int i;

    for (i = size; i > 0; i--) {
	*(((short *)dest)+(i-1)) = xlatepal[*(dest+(i-1))];
    }
}

unsigned char *vbpfunc8(int line)
{
    return cur_buf + (line << 8);
}

unsigned char *vbpfunc16(int line)
{
    return cur_buf + (line << 8);
}

#include "nespal.c"

void init_xlate8(void)
{
    int i;
    XColor color;
    
    vid_pre_xlat = malloc(64);
    vidxlate = NULL;
    video_get_vbp = vbpfunc8;
    
    for (i = 0; i < 64; i++) {
	color.flags = DoRed | DoGreen | DoBlue;
	color.red = nes_palbase_red[i];
	color.green = nes_palbase_green[i];
	color.blue = nes_palbase_blue[i];
	color.red += color.red << 8;
	color.green += color.green << 8;
	color.blue += color.blue << 8;
	if (XAllocColor(display, nes_cmap, &color)) {
	    vid_pre_xlat[i] = color.pixel;
	}
    }
}

void init_xlate16(void)
{
    int i;
    XColor color;

    xlatepal = malloc(128);
    vid_pre_xlat = malloc(64);
    vidxlate = (xlatfunc)&vidxlate16;
    video_get_vbp = vbpfunc16;
    
    for (i = 0; i < 64; i++) {
	vid_pre_xlat[i] = i;
	color.flags = DoRed | DoGreen | DoBlue;
	color.red = nes_palbase_red[i];
	color.green = nes_palbase_green[i];
	color.blue = nes_palbase_blue[i];
	color.red += color.red << 8;
	color.green += color.green << 8;
	color.blue += color.blue << 8;
	if (XAllocColor(display, nes_cmap, &color)) {
	    xlatepal[i] = color.pixel;
	}
    }
}

/*
 * $Log: video_x.c,v $
 * Revision 1.11  1998/08/29 22:09:37  nyef
 * switched to a double-buffered video system. added routines to support
 * the PPU writing directly into the video buffers. eliminated the 8-bit
 * vidxlate routine. changed vidxlate16 to operate using only one buffer.
 * changed the video buffer size to a symbolic constant.
 *
 * Revision 1.10  1998/08/29 14:52:29  nyef
 * added an identity vid_pre_xlat table for use in 16-bitplane modes.
 *
 * Revision 1.9  1998/08/28 03:02:16  nyef
 * added a "vid_pre_xlat" table so that some of the color translation
 * could be done by the PPU. also moved the 8-bit color translation to
 * the PPU and set the 8-bit xlatfinc to be memcpy. (maybe I should
 * double buffer the display?)
 *
 * Revision 1.8  1998/08/15 15:45:34  nyef
 * reworked vidxlate16 to use a normal loop and to copy longs.
 * this should be slightly faster on a pentium.
 *
 * Revision 1.7  1998/08/15 01:08:15  nyef
 * changed unsupported display depth message to indicate current depth.
 *
 * Revision 1.6  1998/08/02 02:03:08  nyef
 * preliminary GUIfication completed.
 *
 * Revision 1.5  1998/08/01 23:12:45  nyef
 * removed the checks for the symbol "GRAPHICS".
 * removed two commented-out printf statements in the input handler.
 *
 * Revision 1.4  1998/08/01 00:57:13  nyef
 * moved old palette values into nespal.c
 *
 * Revision 1.3  1998/07/26 13:22:55  nyef
 * changed vidxlate16 tu use pointer notation and duff's device.
 *
 * Revision 1.2  1998/07/18 18:54:27  nyef
 * changed vidxlate16 to copy shorts instead of chars for a massive speed increase
 *   in 16 bitplane mode.
 *
 * Revision 1.1  1998/07/11 22:19:25  nyef
 * Initial revision
 *
 */
